from unittest import TestCase

import numpy as np

from scidatacontainer import Container, register
from scidatacontainer.fileimage import PngFile

from . import get_test_container


class FileImageTest(TestCase):
    def test_encode_decode(self):
        a = get_test_container()
        register("png", PngFile)
        data = np.random.randint(0, 256, (100, 100, 3))

        a["data/test.png"] = data

        a.write("test.zdc")

        b = Container(file="test.zdc")

        self.assertTrue(np.allclose(a["data/test.png"], b["data/test.png"]))

    def test_hash(self):
        a = get_test_container()
        register("png", PngFile)
        data = np.reshape(np.arange(100 * 100 * 3), (100, 100, 3))

        a["data/test.png"] = data
        a.freeze()

        a.write("test.zdc")

        b = Container(file="test.zdc")
        b.hash()

        self.assertEqual(
            a["content.json"]["hash"],
            b["content.json"]["hash"],
        )
        self.assertEqual(
            b["content.json"]["hash"],
            "bd221ad18daf9d30c7cd83d4edc93be66b87a5796c15ae00c343d6edb70232ba",
        )

    def test_legacy_hash(self):
        a = get_test_container()
        a["content.json"]["modelVersion"] = "1.0.0"
        register("png", PngFile)
        data = np.reshape(np.arange(100 * 100 * 3), (100, 100, 3))

        a["data/test.png"] = data
        a.freeze()

        a.write("test.zdc")

        b = Container(file="test.zdc")
        b.hash()

        self.assertEqual(
            a["content.json"]["hash"],
            b["content.json"]["hash"],
        )
        self.assertEqual(
            b["content.json"]["hash"],
            "d5dd9ea62ac5e3a6e11fa0a5efc60b86338fc27f5e9f9774adbc6b2bd9af8acf",
        )



import os
import unittest

from requests import HTTPError

from ._abstract_container_test import has_connection
from .test_single_step_container import AbstractSingleStepContainerTest


class _TestUpload(AbstractSingleStepContainerTest):
    test_server_url = os.getenv("SCIDATA_TEST_SERVER")
    api_key = os.getenv("SCIDATA_TEST_KEY")
    __test__ = False

    def run(self, result=None):
        if self.__test__:
            return super().run(result)

    def test_upload(self):
        self.test_container_creation()
        uuid = "00000000-0000-0000-0000-000000000000"
        self.dc["content.json"]["uuid"] = uuid
        self.dc.upload(server=self.test_server_url, key=self.api_key)

    def test_upload_bytes(self):
        self.test_container_creation()
        uuid = "00000000-0000-0000-0000-000000000000"
        self.dc["content.json"]["uuid"] = uuid
        b = b"".join(chunk for chunk in self.dc.encode())
        self.dc.upload(data=b, server=self.test_server_url, key=self.api_key)

    def test_no_url(self):
        self.test_container_creation()
        with self.assertRaisesRegex(RuntimeError, "Server URL is missing!"):
            self.dc.upload(server=False, key=self.api_key)

    def test_wrong_url(self):
        self.test_container_creation()
        with self.assertRaisesRegex(
            ConnectionError, "Connection to server " + "http://localhost:54321 failed!"
        ):
            self.dc.upload(server="http://localhost:54321", key=self.api_key)

    def test_no_key(self):
        self.test_container_creation()
        with self.assertRaisesRegex(RuntimeError, "Server API key is missing!"):
            self.dc.upload(key=False)

    def test_wrong_key(self):
        self.test_container_creation()
        with self.assertRaisesRegex(
            HTTPError, f"401 Client Error: Unauthorized for url: {self.test_server_url}"
        ):
            self.dc.upload(server=self.test_server_url, key="abcd")

    def test_invalid_container(self):
        self.test_container_creation()
        del self.dc["content.json"]["uuid"]
        with self.assertRaisesRegex(
            HTTPError, "400 Bad Request: Invalid container " + "content"
        ):
            self.dc.upload(server=self.test_server_url, key=self.api_key)

    def test_no_write_permissions(self):
        self.test_container_creation()
        uuid = "00000000-0000-0000-0000-000000000403"
        self.dc["content.json"]["uuid"] = uuid
        with self.assertRaisesRegex(HTTPError, "403 Forbidden: Unauthorized access"):
            self.dc.upload(server=self.test_server_url, key=self.api_key)

    def test_conflicting_uuid(self):
        self.test_container_creation()
        uuid = "00000000-0000-0000-0000-000000000409"
        self.dc["content.json"]["uuid"] = uuid
        with self.assertRaisesRegex(
            HTTPError, "409 Conflict: UUID is already existing"
        ):
            self.dc.upload(server=self.test_server_url, key=self.api_key)

    @unittest.expectedFailure
    def test_static_already_exist(self, server=None):
        self.test_container_creation()
        uuid = "00000000-0000-0000-0000-000000000301"
        self.dc["content.json"]["uuid"] = uuid
        self.dc.upload(server=self.test_server_url, key=self.api_key)
        self.assertEqual(self.dc["content.json"]["uuid"], uuid[:-3] + "000")


@unittest.skipUnless(has_connection, "No server conntection available!")
class TestUpload(_TestUpload):
    __test__ = True



##########################################################################
# Copyright (c) 2023 Reinhard Caspary                                    #
# <reinhard.caspary@phoenixd.uni-hannover.de>                            #
# This program is free software under the terms of the MIT license.      #
##########################################################################
#
# This module provides data conversion classes for image files. All
# data conversion classes should inherit from filebase.AbstractFile and must
# provide three methods:
#
# encode(): Return data encoded as bytes string.
# decode(data): Decode and store given bytes string data.
# hash(): Return SHA256 hash from data as hex string.
#
# The hash implementation shoud make sure that semantically equivalent
# data results in the same hash.
#
##########################################################################

import cv2 as cv
import numpy as np

from .filebase import AbstractFile


class PngFile(AbstractFile):

    """ Data conversion class for PNG image. """

    def encode(self):

        """ Convert image to bytes string. """

        return cv.imencode('.png', self.data)[1].tobytes()

    def decode(self, data):

        """ Decode image from bytes string. """

        flags = cv.IMREAD_ANYDEPTH | cv.IMREAD_ANYCOLOR
        self.data = np.frombuffer(data, dtype=np.uint8)
        self.data = cv.imdecode(self.data, flags=flags)


register = [
    ("png", PngFile, None),
    ]



import copy
import os
import uuid

from scidatacontainer import Container

from ._abstract_container_test import AbstractContainerTest


class AbstractSingleStepContainerTest(AbstractContainerTest):
    def _compare_with_items(self, dc):
        super()._compare_with_items(dc)
        self.assertFalse(dc["content.json"]["static"])
        self._check_timestamp(dc["content.json"]["created"])
        self._check_timestamp(dc["content.json"]["storageTime"])
        self.assertTrue(dc["content.json"]["complete"])

    def _compare_with_minimal_items(self, dc):
        super()._compare_with_minimal_items(dc)
        self.assertFalse(dc["content.json"]["static"])
        self._check_timestamp(dc["content.json"]["created"])
        self._check_timestamp(dc["content.json"]["storageTime"])
        self.assertTrue(dc["content.json"]["complete"])

    def test_container_creation(self):
        items = self.items
        with self.assertRaisesRegex(RuntimeError, "No data!"):
            dc = Container()

        dc = Container(items=items)

        self._compare_with_items(dc)

        self.assertIsNone(dc["content.json"]["hash"])
        dc.hash()
        self.assertIsNotNone(dc["content.json"]["hash"])

        self.dc = dc


class TestSingleStepContainer(AbstractSingleStepContainerTest):
    def test_container_creation_minimal(self):
        items = self.items_minimal
        dc = Container(items=items)

        self._compare_with_minimal_items(dc)

        self.dc = dc

    def test_write(self, clean=True):
        self.test_container_creation()
        self.dc.write(self.export_filename)
        if clean:
            os.remove(self.export_filename)

    def test_write_bytes(self, clean=True):
        self.test_container_creation()
        self.dc.write(self.export_filename, b"".join(self.dc.encode()))
        dc = Container(file=self.export_filename)

        self._compare_with_items(dc)

        self.assertIsNotNone(dc["content.json"]["hash"])
        old_hash = dc["content.json"]["hash"]

        dc.hash()
        self.assertIsNotNone(dc["content.json"]["hash"])
        # new hash is equal with the old one
        self.assertEqual(old_hash, dc["content.json"]["hash"])
        if clean:
            os.remove(self.export_filename)

    def test_read(self):
        self.test_write(clean=False)
        dc = Container(file=self.export_filename)

        self._compare_with_items(dc)

        self.assertIsNotNone(dc["content.json"]["hash"])
        old_hash = dc["content.json"]["hash"]

        dc.hash()
        self.assertIsNotNone(dc["content.json"]["hash"])
        # new hash is equal with the old one
        self.assertEqual(old_hash, dc["content.json"]["hash"])

    def test_content_validation(self):
        items = copy.deepcopy(self.items)
        del items["content.json"]
        with self.assertRaisesRegex(RuntimeError, "Item 'content.json' is missing!"):
            Container(items=items)

        items = copy.deepcopy(self.items)
        del items["content.json"]["containerType"]

        with self.assertRaisesRegex(
            RuntimeError, "Attribute 'containerType' is missing!"
        ):
            Container(items=items)

        items = copy.deepcopy(self.items)
        items["content.json"]["containerType"] = ["name", "test"]

        with self.assertRaisesRegex(
            RuntimeError, "Attribute containerType is no " + "dictionary!"
        ):
            Container(items=items)

        items = copy.deepcopy(self.items)
        del items["content.json"]["containerType"]["name"]
        with self.assertRaisesRegex(RuntimeError, "Name of containerType is missing!"):
            Container(items=items)

        items = copy.deepcopy(self.items)
        del items["content.json"]["containerType"]["version"]
        with self.assertRaisesRegex(
            RuntimeError, "Version of containerType is missing!"
        ):
            Container(items=items)

        items = copy.deepcopy(self.items)
        del items["content.json"]["usedSoftware"][0]["name"]
        with self.assertRaisesRegex(RuntimeError, "Software name is missing!"):
            Container(items=items)

        items = copy.deepcopy(self.items)
        del items["content.json"]["usedSoftware"][0]["version"]
        with self.assertRaisesRegex(RuntimeError, "Software version is missing!"):
            Container(items=items)

        items = copy.deepcopy(self.items)
        del items["content.json"]["usedSoftware"][1]["idType"]
        with self.assertRaisesRegex(
            RuntimeError, "Type of software reference id is " + "missing!"
        ):
            Container(items=items)

    def test_meta_validation(self):
        items = copy.deepcopy(self.items)
        del items["meta.json"]
        with self.assertRaisesRegex(RuntimeError, "Item 'meta.json' is missing!"):
            Container(items=items)

        items = copy.deepcopy(self.items)
        del items["meta.json"]["title"]
        with self.assertRaisesRegex(RuntimeError, "Data title is missing!"):
            Container(items=items)

    def test_mutability(self):
        self.test_container_creation()
        self.assertFalse(self.dc.mutable)
        self.dc.release()
        self.assertTrue(self.dc.mutable)
        self.dc.release()
        self.assertTrue(self.dc.mutable)

    def test_print(self):
        self.test_container_creation()
        s = self.dc.__str__()

        self.assertIn("Complete Container", s)
        ct = self.items["content.json"]["containerType"]
        self.assertIn(
            "type:        " + ct["name"] + " " + ct["version"] + " (" + ct["id"] + ")",
            s,
        )
        self.assertIn("uuid:        " + self.dc["content.json"]["uuid"], s)
        self.assertIn("replaces:    " + self.dc["content.json"]["replaces"], s)
        self.assertIn("created:     " + self.dc["content.json"]["created"], s)
        self.assertIn("storageTime: " + self.dc["content.json"]["storageTime"], s)
        self.assertIn("author:      " + self.dc["meta.json"]["author"], s)

    def test_setitem(self):
        self.test_container_creation()

        with self.assertRaisesRegex(RuntimeError, "Immutable container!"):
            self.dc["data/test.txt"] = "Test123"

        self.dc.release()
        self.dc["data/test.txt"] = "Test123"
        self.dc["data/test.bin"] = b"Test123"
        self.dc["data/test.abc"] = b"Test123"

        with self.assertRaisesRegex(
            RuntimeError,
            "No matching file format found for item " + "'data/test.list'!",
        ):
            self.dc["data/test.list"] = ["Test123", "Test456"]

    def test_delitem(self):
        self.test_container_creation()
        with self.assertRaisesRegex(RuntimeError, "Immutable container!"):
            del self.dc["data/parameter.json"]

        self.dc.release()
        del self.dc["data/parameter.json"]

    def test_getitem(self):
        self.test_container_creation()
        self._check_timestamp(self.dc["content.json"]["created"])

        with self.assertRaisesRegex(KeyError, "Unknown item 'data/test.abc'!"):
            self.dc["data/test.abc"]

    def test_values(self):
        self.test_container_creation()
        self.assertEqual(
            self.dc.values(),
            [self.dc["content.json"], self.parameter, self.data, self.dc["meta.json"]],
        )

    def test_content_getter(self):
        self.test_container_creation()
        self.assertDictEqual(self.dc.content, self.dc["content.json"])
        self.assertTrue(self.dc["content.json"]["complete"])
        self.dc.content["complete"] = False
        self.assertFalse(self.dc["content.json"]["complete"])
        self.assertDictEqual(self.dc.content, self.dc["content.json"])

    def test_content_setter(self):
        self.test_container_creation()
        self.assertDictEqual(self.dc.content, self.dc["content.json"])
        self.dc.release()
        d2 = {"test123": "test"}
        self.dc.content = d2
        self.assertDictEqual(self.dc.content, d2)
        self.assertDictEqual(self.dc["content.json"], d2)

    def test_meta_getter(self):
        self.test_container_creation()
        self.assertDictEqual(self.dc.meta, self.dc["meta.json"])
        self.assertEqual(
            self.dc["meta.json"]["title"], "This is a sample image dataset"
        )

        self.dc.meta["title"] = "This is NOT a sample image dataset"
        self.assertEqual(
            self.dc["meta.json"]["title"], "This is NOT a sample image dataset"
        )
        self.assertDictEqual(self.dc.meta, self.dc["meta.json"])

    def test_meta_setter(self):
        self.test_container_creation()
        self.assertDictEqual(self.dc.meta, self.dc["meta.json"])
        self.dc.release()
        d2 = {"test123": "test"}
        self.dc.meta = d2
        self.assertDictEqual(self.dc.meta, d2)
        self.assertDictEqual(self.dc["meta.json"], d2)

    def test_uuid_getter(self):
        self.test_container_creation()
        dc_uuid = self.dc["content.json"]["uuid"]
        self.assertEqual(dc_uuid, self.dc.uuid)

        self.dc["content.json"]["uuid"] = str(uuid.uuid4())
        self.assertNotEqual(dc_uuid, self.dc.uuid)

    def test_uuid_setter(self):
        self.test_container_creation()
        with self.assertRaises(AttributeError):
            self.dc.uuid = str(uuid.uuid4())



import copy
import os

from scidatacontainer import Container

from ._abstract_container_test import AbstractContainerTest


class TestORCID(AbstractContainerTest):
    def test_valid_orcid(self):
        orcid = "0000-0002-9079-593X"
        items = copy.deepcopy(self.items)
        dc = Container(items=items, config={"orcid": orcid})

        self._compare_with_items(dc)
        self.assertEqual(dc.meta["orcid"], orcid)

    def test_format_orcid(self):
        orcid = "0000-0002-9079-593X"
        items = copy.deepcopy(self.items)
        dc = Container(
            items=items, config={"orcid": orcid.replace("-", "").lower()}
        )

        self._compare_with_items(dc)
        self.assertEqual(dc.meta["orcid"], orcid)

    def test_length_orcid(self):
        orcid = "00-0002-9079-593X"
        items = copy.deepcopy(self.items)
        dc = Container(
            items=items, config={"orcid": orcid.replace("-", "").lower()}
        )

        self._compare_with_items(dc)
        self.assertEqual(dc.meta["orcid"], "")

    def test_num_space_orcid(self):
        for orcid in [
            "0000-0001-5000-0007",
            "0000-0003-5000-0001",
            "0009-0000-0000-0009",
            "0009-0010-0000-0003",
        ]:
            items = copy.deepcopy(self.items)
            dc = Container(items=items, config={"orcid": orcid})
            self._compare_with_items(dc)
            self.assertEqual(dc.meta["orcid"], orcid)

        # Check first ORCID outside the range of valid numbers
        for orcid in [
            "0000-0001-4999-9992",
            "0000-0003-5000-001X",
            "0008-9999-9999-9996",
            "0009-0010-0000-0011",
        ]:
            items = copy.deepcopy(self.items)
            dc = Container(items=items, config={"orcid": orcid})
            self._compare_with_items(dc)
            self.assertEqual(dc.meta["orcid"], "")

    def test_invalid_orcid(self):
        orcid = "0000-0001-5000-0X07"
        items = copy.deepcopy(self.items)
        dc = Container(items=items, config={"orcid": orcid})
        self._compare_with_items(dc)
        self.assertEqual(dc.meta["orcid"], "")

    def test_invalid_checksum(self):
        orcid = "0000-0002-9079-5930"
        items = copy.deepcopy(self.items)
        dc = Container(items=items, config={"orcid": orcid})
        self._compare_with_items(dc)
        self.assertEqual(dc.meta["orcid"], "")

    def test_write_read(self, clean=True):
        orcid = "0000-0002-9079-593X"
        items = copy.deepcopy(self.items)
        dc = Container(items=items, config={"orcid": orcid})
        self._compare_with_items(dc)
        self.assertEqual(dc.meta["orcid"], orcid)
        dc.write(self.export_filename)

        dc2 = Container(file=self.export_filename)
        self._compare_with_items(dc2)
        self.assertEqual(dc2.meta["orcid"], orcid)

        if clean:
            os.remove(self.export_filename)



##########################################################################
# Copyright (c) 2023 Reinhard Caspary                                    #
# <reinhard.caspary@phoenixd.uni-hannover.de>                            #
# This program is free software under the terms of the MIT license.      #
##########################################################################
#
# This module provides data conversion classes for image files. All
# data conversion classes should inherit from filebase.AbstractFile and must
# provide three methods:
#
# encode(): Return data encoded as bytes string.
# decode(data): Decode and store given bytes string data.
# hash(): Return SHA256 hash from data as hex string.
#
# The hash implementation shoud make sure that semantically equivalent
# data results in the same hash.
#
##########################################################################

import io

import numpy as np

from .filebase import AbstractFile


class NpyFile(AbstractFile):
    """Data conversion class for NumPy arrays (ndarray)."""

    allow_pickle = False

    def encode(self):
        """Convert NumPy array to bytes string."""

        with io.BytesIO() as fp:
            np.save(fp, self.data, allow_pickle=self.allow_pickle)
            fp.seek(0)
            data = fp.read()
        return data

    def decode(self, data):
        """Decode NumPy array from bytes string."""

        with io.BytesIO() as fp:
            fp.write(data)
            fp.seek(0)
            self.data = np.load(fp, allow_pickle=self.allow_pickle)


register = [
    ("npy", NpyFile, np.ndarray),
]



import io
from hashlib import sha256

import h5py
import numpy as np

from scidatacontainer import AbstractFile


class Hdf5File(AbstractFile):
    def encode(self):
        with io.BytesIO() as fp:
            with h5py.File(fp, "w") as h5file:
                if isinstance(self.data, np.ndarray):
                    _ = h5file.create_dataset(
                        "dataset",
                        data=self.data,
                        compression="gzip",
                        compression_opts=9,
                    )
                elif isinstance(self.data, dict):
                    for key in sorted(self.data.keys()):
                        assert key != "dataset"
                        value = self.data[key]
                        if isinstance(value, (np.ndarray, h5py.Dataset)):
                            _ = h5file.create_dataset(
                                key, data=value, compression="gzip", compression_opts=9
                            )
                        else:
                            h5file.attrs[key] = value
                else:
                    raise NotImplementedError(
                        f"Unknown data type {type(self.data).__name__}!"
                    )
            _ = fp.seek(0)
            data = fp.read()
        return data

    def decode(self, data):
        with io.BytesIO(data) as fp:
            with h5py.File(fp, "r") as h5file:
                datasets = [
                    name
                    for name in h5file.keys()
                    if isinstance(h5file[name], h5py.Dataset)
                ]
                if datasets == ["dataset"]:
                    self.data = h5file["dataset"][()]
                else:
                    self.data = {name: h5file[name][()] for name in datasets}
                    self.data.update(
                        {
                            name: getattr(
                                h5file.attrs[name], "tolist", lambda: h5file.attrs[name]
                            )()
                            for name in h5file.attrs.keys()
                        }
                    )

    def hash(self) -> str:
        if isinstance(self.data, np.ndarray):
            return sha256(self.data.data).hexdigest()

        return sha256(
            " ".join(
                [
                    " ".join(
                        [
                            sha256(key.encode("utf-8")).hexdigest(),
                            sha256(
                                getattr(
                                    self.data[key],
                                    "data",
                                    np.array(self.data[key]).data,
                                )
                            ).hexdigest(),
                        ]
                    )
                    for key in sorted(self.data.keys())
                ]
            ).encode("utf-8")
        ).hexdigest()


register = [("hdf5", Hdf5File, np.ndarray)]



##########################################################################
# Copyright (c) 2023-2024 Reinhard Caspary                               #
# <reinhard.caspary@phoenixd.uni-hannover.de>                            #
# This program is free software under the terms of the MIT license.      #
##########################################################################
#
# This package provides the Scientific Data Container as class Container
# which may be stored as a ZIP package containing items (files). Based
# on their file extension, the following item types are supported:
#
# .json: JSON file
# .txt:  Encoded text file (default encoding: UTF-8)
# .log:  Encoded text file (default encoding: UTF-8)
# .pgm:  Encoded text file (default encoding: UTF-8)
# .bin:  Raw binary data file
# .npy:  NumPy array (requires Python module numpy)
# .png:  PNG image (requires Python modules cv2 and numpy)
#
# Users may register other file extensions to file conversion classes
# using the function register(). See package fileimage as an example for
# such a conversion class.
#
##########################################################################

__all__ = [
    "timestamp",
    "modelVersion",
    "load_config",
    "register",
    "Container",
]

import typing
from importlib import import_module

from .config import load_config
from .container import MODELVERSION as modelVersion
from .container import AbstractContainer, timestamp
from .filebase import AbstractFile

__version__ = "1.2.0"

suffixes = {}
classes = {}
formats = []


def register(
    suffix: str,
    fclass: typing.Type[AbstractFile],
    pclass: typing.Type[object] = None,
):
    """Register a suffix to a conversion class.

    If the parameter class is a string, it is interpreted as known suffix and
    the conversion class of this suffix is registered also for the new one.

    Args:
        suffix: file suffix to identify this file type.
        fclass: Conversion class derived from AbstractFile.
        pclass: Python class that represents this object type.
    """

    if isinstance(fclass, str):
        if pclass is not None:
            raise RuntimeError("Alias %s:%s with default class!" % (suffix, fclass))
        fclass = suffixes[fclass]

    # Simple sanity check for the class interface
    for method in ("encode", "decode", "hash"):
        if not hasattr(fclass, method) or not callable(getattr(fclass, method)):
            raise RuntimeError(
                "No method %s() in class for suffix '%s'!" % (method, suffix)
            )

    # Register suffix
    suffixes[suffix] = fclass

    # Register Python class. Last registration becomes default.
    # Overriding the mapping dict:JsonFile is not allowed.
    if pclass not in classes or pclass is not dict:
        classes[pclass] = fclass

    # Register unknown file format
    if fclass not in formats:
        formats.append(fclass)


# Initialize the conversion class database

for name in ("filebase", "fileimage", "filenumpy", "filehdf5"):
    fullname = __name__ + "." + name
    try:
        module = import_module(fullname)
    except ModuleNotFoundError:
        continue

    for suffix, fclass, pclass in module.register:
        register(suffix, fclass, pclass)


# Inject certain known file formats into the container class
class Container(AbstractContainer):
    """Scientific data container."""

    _suffixes = suffixes
    _classes = classes
    _formats = formats



import copy
from datetime import datetime, timedelta
import os

from scidatacontainer import Container

from ._abstract_container_test import AbstractContainerTest


class TestMultiStepContainer(AbstractContainerTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.items["content.json"]["static"] = False
        cls.items["content.json"]["complete"] = False
        cls.items_minimal["content.json"]["complete"] = False
        created = datetime.fromisoformat(cls.items["content.json"]["created"])
        created -= timedelta(seconds=1)
        cls.created = created.isoformat(timespec="seconds")
        cls.items["content.json"]["created"] = cls.created
        cls.items_minimal["content.json"]["created"] = cls.created

    def _compare_with_items(self, dc):
        super()._compare_with_items(dc)
        self._check_timestamp(dc["content.json"]["created"])

        self._check_timestamp(dc["content.json"]["storageTime"])

        self.assertFalse(dc["content.json"]["complete"])

        self.assertFalse(dc["content.json"]["static"])

    def _compare_with_minimal_items(self, dc):
        super()._compare_with_minimal_items(dc)
        self._check_timestamp(dc["content.json"]["created"])

        self._check_timestamp(dc["content.json"]["storageTime"])

        self.assertFalse(dc["content.json"]["complete"])

        self.assertFalse(dc["content.json"]["static"])

    def test_container_creation(self):
        items = copy.deepcopy(self.items)
        dc = Container(items=items)

        self._compare_with_items(dc)

        self.dc = dc

    def test_container_creation_minimal(self):
        items = copy.deepcopy(self.items_minimal)
        dc = Container(items=items)

        self._compare_with_minimal_items(dc)

        self.dc = dc

    def test_write(self, clean=True):
        self.test_container_creation()
        self.dc.write(self.export_filename)
        if clean:
            os.remove(self.export_filename)

    def test_read(self):
        self.test_write(clean=False)
        dc = Container(file=self.export_filename)

        self._compare_with_items(dc)

    def test_print(self):
        self.test_container_creation()
        s = self.dc.__str__()

        self.assertIn("Incomplete Container", s)
        ct = self.items["content.json"]["containerType"]
        self.assertIn("type:        " + ct["name"] + " " + ct["version"]
                      + " (" + ct["id"] + ")", s)
        self.assertIn("uuid:        " + self.dc["content.json"]["uuid"], s)
        self.assertIn("replaces:    " + self.dc["content.json"]["replaces"], s)
        self.assertIn("created:     " + self.dc["content.json"]["created"], s)
        self.assertIn("storageTime: " + self.dc["content.json"]["storageTime"],
                      s)
        self.assertIn("author:      " + self.dc["meta.json"]["author"], s)

        self.dc["content.json"]["complete"] = True

        s = self.dc.__str__()

        self.assertIn("Complete Container", s)
        ct = self.items["content.json"]["containerType"]
        self.assertIn("type:        " + ct["name"] + " " + ct["version"]
                      + " (" + ct["id"] + ")", s)
        self.assertIn("uuid:        " + self.dc["content.json"]["uuid"], s)
        self.assertIn("replaces:    " + self.dc["content.json"]["replaces"], s)
        self.assertIn("created:     " + self.dc["content.json"]["created"], s)
        self.assertIn("storageTime: " + self.dc["content.json"]["storageTime"],
                      s)
        self.assertIn("author:      " + self.dc["meta.json"]["author"], s)



from unittest import TestCase

from jsonschema.exceptions import ValidationError

from scidatacontainer.jsonschema import VERSIONS_AVAILABLE, content, meta, validate
from scidatacontainer.tests import get_test_container


class JSONSchemaTest(TestCase):
    def test_content(self):
        self.assertIsInstance(content, dict)

        self.assertEqual(len(VERSIONS_AVAILABLE), len(content))

        for v in VERSIONS_AVAILABLE:
            self.assertIn(v, content.keys())

    def test_meta(self):
        self.assertIsInstance(meta, dict)

        self.assertEqual(len(VERSIONS_AVAILABLE), len(meta))

        for v in VERSIONS_AVAILABLE:
            self.assertIn(v, meta.keys())

    def test_validate_content(self):
        a = get_test_container()
        validate(
            a["content.json"],
            schema=content[VERSIONS_AVAILABLE[-1]],
            schema_name="content",
        )

    def test_validate_meta(self):
        a = get_test_container()
        validate(
            a["meta.json"], schema=meta[VERSIONS_AVAILABLE[-1]], schema_name="meta"
        )

    def test_validate_missing_attribute(self):
        a = get_test_container()
        del a["content.json"]["complete"]
        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
            )

        self.assertEqual(
            cm.exception.message, "'complete' is a required property in content.json."
        )

        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
                translate=False,
            )

        self.assertEqual(cm.exception.message, "'complete' is a required property")

        a = get_test_container()
        del a["meta.json"]["author"]
        with self.assertRaises(ValidationError) as cm:
            validate(
                a["meta.json"], schema=meta[VERSIONS_AVAILABLE[-1]], schema_name="meta"
            )

        self.assertEqual(
            cm.exception.message, "'author' is a required property in meta.json."
        )

    def test_validate_wrong_type(self):
        a = get_test_container()
        a["content.json"]["complete"] = "True"
        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
            )

        self.assertEqual(
            cm.exception.message,
            "Value of 'complete' in content.json has the wrong "
            + "type: 'True' is not of type 'boolean'.",
        )

        a = get_test_container()
        a["content.json"]["created"] = 42
        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
            )

        self.assertEqual(
            cm.exception.message,
            "Value of 'created' in content.json has the wrong "
            + "type: 42 is not of type 'string'.",
        )

    def test_validate_wrong_format(self):
        a = get_test_container()
        a["content.json"]["uuid"] = "test123"
        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
            )

        self.assertEqual(
            cm.exception.message,
            "Value of 'uuid' in content.json has the wrong "
            + "format: 'test123' is not a 'uuid'.",
        )

        a = get_test_container()
        a["content.json"]["created"] = "test123"
        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
            )

        self.assertEqual(
            cm.exception.message,
            "Value of 'created' in content.json has the wrong "
            + "format: 'test123' is not a 'date-time'.",
        )

        validate(
            a["content.json"],
            schema=content[VERSIONS_AVAILABLE[-1]],
            schema_name="content",
            check_format=False,
        )

        a = get_test_container()
        a["meta.json"]["email"] = "test123"
        with self.assertRaises(ValidationError) as cm:
            validate(
                a["meta.json"], schema=meta[VERSIONS_AVAILABLE[-1]], schema_name="meta"
            )

        self.assertEqual(
            cm.exception.message,
            "Value of 'email' in meta.json has the wrong "
            + "format: 'test123' is not a 'email'.",
        )

        a = get_test_container()
        a["content.json"]["hash"] = "test"
        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
            )

        self.assertEqual(
            cm.exception.message,
            "Value of 'hash' in content.json has the wrong "
            + "format: A hash can only contain hex digits (0-9, "
            + "a-f and A-F).",
        )

        a = get_test_container()
        a["content.json"]["modelVersion"] = "1.bcd.2.3"
        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
            )

        self.assertEqual(
            cm.exception.message,
            "Value of 'modelVersion' in content.json has the "
            + "wrong format: A model version only contains digits"
            + " and dots.",
        )

    def test_guess_schema(self):
        a = get_test_container()
        a["content.json"]["uuid"] = "test123"
        with self.assertRaises(ValidationError) as cm:
            validate(a["content.json"])

        self.assertEqual(
            cm.exception.message,
            "Value of 'uuid' in content.json has the wrong "
            + "format: 'test123' is not a 'uuid'.",
        )

        a["meta.json"]["timestamp"] = "test123"
        with self.assertRaises(ValidationError) as cm:
            validate(a["meta.json"])

        self.assertEqual(
            cm.exception.message,
            "Value of 'timestamp' in meta.json has the wrong "
            + "format: 'test123' is not a 'date-time'.",
        )

    def test_static_requires_hash(self):
        a = get_test_container()
        a.freeze()
        validate(
            a["content.json"],
            schema=content[VERSIONS_AVAILABLE[-1]],
            schema_name="content",
        )

        del a["content.json"]["hash"]

        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
            )

        self.assertEqual(cm.exception.message, "A static container requires a hash.")

        a["content.json"]["hash"] = None

        with self.assertRaises(ValidationError) as cm:
            validate(
                a["content.json"],
                schema=content[VERSIONS_AVAILABLE[-1]],
                schema_name="content",
            )

        self.assertEqual(cm.exception.message, "A static container requires a hash.")



##########################################################################
# Copyright (c) 2023-2024 Reinhard Caspary                               #
# <reinhard.caspary@phoenixd.uni-hannover.de>                            #
# This program is free software under the terms of the MIT license.      #
##########################################################################
#
# This module provides the Scientific Data Container as class
# DataContainer which may be stored or uploaded as a ZIP package
# containing items (files). Do not use this class directly! Use the
# class Container provided by the package scidatacontainer instead.
#
##########################################################################

import copy
import hashlib
import io
import json
import pathlib
import typing
import uuid
from abc import ABC
from collections.abc import Iterator
from datetime import datetime, timezone
from queue import Queue
from threading import Thread
from types import SimpleNamespace
from zipfile import ZIP_DEFLATED, ZipFile

import requests

from .config import load_config
from .filebase import BinaryFile, JsonFile, TextFile

# Version of the implemented data model
MODELVERSION = "1.0.1"


##########################################################################
# Timestamp function


def timestamp() -> str:
    """Return the current ISO 8601 compatible timestamp as string.

    Returns:
        str: timestamp as string
    """
    return datetime.now(timezone.utc).isoformat(timespec="seconds")


##########################################################################
# Data container class


class AbstractContainer(ABC):
    """Scientific data container with minimal file support.

    The following file types are supported:
        - .json <-> dict
        - .txt <-> str
        - .bin <-> bytes
    """

    _config = None
    _suffixes = {"json": JsonFile, "txt": TextFile, "bin": BinaryFile}
    _classes = {dict: JsonFile, str: TextFile, bytes: BinaryFile}
    _formats = [TextFile]

    def __init__(
        self,
        items: dict | None = None,
        file: str | None = None,
        uuid: str | None = None,
        config: dict | None = None,
        compression: int = ZIP_DEFLATED,
        compresslevel: int = -1,
        ignore_items: list[str] = [],
        **kwargs,
    ):
        """Construct a DataContainer object.

        It will try the following in the specified order:
            - Create a new DataContainer if items is passed as argument.
            - Load local DataContainer from hard drive if file is passed.
            - Download a DataContainer from a server if uuid is passed.

        Args:
            items: Dictionary of items to build a new DataContainer.
            file: Filename to read a DataContainer from local hard drive.
            uuid: UUID of a Container to download from a server instance.
            author: Author name for meta.json
            email: Author email for meta.json
            server: URL of the server instance providing the Container.
            key: API-Key from the server to identify yourself.
            compression: Numeric constant for the compression method
            compresslevel: Level of compression, 0-fastest, 9-best compression
            ignore_items: List of container items that are not loaded to memory
        """
        self.kwargs = {
            "items": items,
            "file": file,
            "uuid": uuid,
            "config": config,
            "compression": compression,
            "compresslevel": compresslevel,
            "ignore_items": ignore_items,
        }
        self.kwargs.update(kwargs)

        self._items = {}
        self.__pre_init__()

        # Load variables from kwargs in namespace
        n = SimpleNamespace(**self.kwargs)

        # Container must be mutable initially
        self.mutable = True

        # Load configuration
        if n.config is not None:
            self._config = dict(n.config)
        else:
            self._config = load_config()

        # Store all items in the container
        if n.items is not None:
            self._store(n.items, True, False)
            self.mutable = not self["content.json"]["static"]

        # Load local container file
        elif n.file is not None:
            self._read(fn=n.file, ignore_items=n.ignore_items)

        # Download container from server
        elif n.uuid is not None:
            server = self._config["server"]
            key = self._config["key"]
            self._download(
                uuid=n.uuid, server=server, key=key, ignore_items=n.ignore_items
            )

        # No data source
        else:
            raise RuntimeError("No data!")

        self.compression = n.compression
        self.compresslevel = n.compresslevel
        self.ignore_items = n.ignore_items

        # Check validity of author ORCID
        self._norm_orcid()

        if self.mutable and len(self.ignore_items) > 0:
            raise RuntimeError(
                "Partial loading of items only supported for immutable containers!"
            )
        self.__post_init__()

    def __pre_init__(self):
        """Manipulate the container before initialization.

        This method can be overwritten by inheriting classes to manipulate the
        container object or the arguments passed to __init__ before the
        arguments are processed in __init__. To do so, manipulate the
        dictionary self.kwargs inside this function.
        """

    def __post_init__(self):
        """Manipulate the container after initialization.

        This method can be overwritten by inheriting classes to manipulate the
        container object after __init__ is executed. This might be useful to
        implement custom checks after creation of a container.
        """

    def _store(self, items, validate=True, strict=True):
        """Store all items in the container."""
        # Add all items in the container
        self._items = {}
        mutable = self.mutable
        self.mutable = True
        for path, data in items.items():
            self[path] = data

        # Make sure that the items content.json and meta.json exist and
        # contain all required attributes
        if "content.json" not in self:
            raise RuntimeError("Item 'content.json' is missing!")
        if "meta.json" not in self:
            raise RuntimeError("Item 'meta.json' is missing!")
        if validate:
            self.validate_content()
            self.validate_meta()

        # validate hash for old modelVersions
        if (
            strict
            and self["content.json"]["hash"]
            and self["content.json"]["modelVersion"] < "1.0.1"
        ):
            old_hash = self["content.json"]["hash"]
            self.hash()
            if old_hash != self["content.json"]["hash"]:
                raise RuntimeError("Wrong hash!")

        if self["content.json"]["static"] and not self["content.json"]["hash"]:
            self.hash()

        # Restore mutable flag
        self.mutable = mutable

    @property
    def content(self) -> dict:
        return self["content.json"]

    @content.setter
    def content(self, value: dict):
        self["content.json"] = value

    @property
    def meta(self) -> dict:
        return self["meta.json"]

    @meta.setter
    def meta(self, value: dict):
        self["meta.json"] = value

    @property
    def uuid(self) -> str:
        return self["content.json"]["uuid"]

    def __contains__(self, path):
        """Return true, if the given path matches an item in this container."""
        return path in self._items

    def __setitem__(self, path, data):
        """Store data as a container item."""
        # Immutable container must not be modified
        if not self.mutable:
            raise RuntimeError("Immutable container!")

        # Store dict with pathlib.Path object as specil types.
        if isinstance(data, dict) and isinstance(data.get("path", None), pathlib.Path):
            self._items[path] = _OnDiskFile(**data)
            return

        # Get file extension
        ext = path.rsplit(".", 1)[1]

        # Unregistered file extension
        if ext not in self._suffixes:
            # Try to convert bytes. Default is BinaryFile.
            if isinstance(data, bytes):
                for cls in self._formats:
                    try:
                        item = cls(data)
                        break
                    except Exception:
                        pass
                else:
                    item = BinaryFile(data)

            # Other Python object must be registered
            else:
                if type(data) in self._classes:
                    cls = self._classes[type(data)]
                    item = cls(data)
                else:
                    raise RuntimeError(
                        "No matching file format found for " + "item '%s'!" % path
                    )

        # Registered file extension
        else:
            cls = self._suffixes[ext]
            item = cls(data)

        # Store conversion object containing data
        self._items[path] = item

    def __getitem__(self, path):
        """Get the data content of a container item."""
        if path in self:
            if isinstance(self._items[path], _OnDiskFile):
                return self._items[path]
            return self._items[path].data
        if path in self.ignore_items:
            raise KeyError(f"Item '{path}' was ignored while reading the file.")
        raise KeyError("Unknown item '%s'!" % path)

    def validate_content(self):
        """Make sure that the item "content.json" exists and contains
        all required attributes.
        """
        # Get a copy of the item "content.json"
        content = copy.deepcopy(self["content.json"])

        # Keep UUID of a multi-step container and create a new one otherwise
        if "uuid" not in content or not content["uuid"]:
            content["uuid"] = str(uuid.uuid4())

        # The optional attribute 'replace' contains the UUID of the
        # predecessor of this container. It replaces the former one,
        # which must have the same containerType and owner and a smaller
        # or equal creation time. The replacement feature should only be
        # used for minor data modifications (e.g. additional keywords or
        # comment in meta.json). The server returns always the latest
        # version.
        if "replaces" not in content:
            content["replaces"] = None

        # The attribute 'containerType' is a dictionary which must at
        # least contain the type of the container as short string
        # without spaces. If the container type is standardized, it must
        # also contain a type id and a version string.
        if "containerType" not in content:
            raise RuntimeError("Attribute 'containerType' is missing!")
        ptype = content["containerType"]
        if not isinstance(ptype, dict):
            raise RuntimeError("Attribute containerType is no dictionary!")
        if "name" not in ptype:
            raise RuntimeError("Name of containerType is missing!")
        if "id" in ptype and "version" not in ptype:
            raise RuntimeError("Version of containerType is missing!")

        # The boolean attribute 'static' is required. Default is False.
        if "static" in content:
            content["static"] = bool(content["static"])
        else:
            content["static"] = False

        # The boolean attribute 'complete' is required. Default is True.
        if not content["static"] and "complete" in content:
            content["complete"] = bool(content["complete"])
        else:
            content["complete"] = True

        # Current time
        ts = timestamp()

        # The attribute 'created' is required. It is created
        # automatically for a new dataset.
        if "created" not in content or not content["created"]:
            content["created"] = ts

        # The attribute 'storageTime' is updated automatically when the
        # container is stored
        if "storageTime" not in content or not content["complete"]:
            content["storageTime"] = ts

        # The attribute 'hash' is optional
        if "hash" not in content or not content["hash"]:
            content["hash"] = None

        # The attribute 'usedSoftware' is a list of dictionaries, which
        # may be empty. Each dictionary must contain atleast the items
        # "name" and "version" specifying name and version of a
        # software. It may also contain the items "id" and "idType"
        # specifying a reference id (e.g. GitHub-URL) and its type.
        if "usedSoftware" not in content or not content["usedSoftware"]:
            content["usedSoftware"] = []
        for sw in content["usedSoftware"]:
            if "name" not in sw:
                raise RuntimeError("Software name is missing!")
            if "version" not in sw:
                raise RuntimeError("Software version is missing!")
            if "id" in sw and "idType" not in sw:
                raise RuntimeError("Type of software reference id is missing!")

        # Version of the data model provided by this package
        content["modelVersion"] = MODELVERSION

        # Store the item "content.json"
        self["content.json"] = content

    def validate_meta(self):
        """Make sure that the item "meta.json" exists and contains all
        required attributes."""
        # Get a copy of the item "meta.json"
        meta = copy.deepcopy(self["meta.json"])

        # Author name is required
        if "author" not in meta:
            meta["author"] = self._config["author"]
        if not meta["author"]:
            raise RuntimeError("Author name is missing!")

        # Author email address is required
        if "email" not in meta:
            meta["email"] = self._config["email"]

        # Author ORCiD is optional
        if "orcid" not in meta:
            meta["orcid"] = self._config["orcid"]

        # Author affiliation is optional
        if "organization" not in meta:
            meta["organization"] = self._config["organization"]

        # Comment on dataset is optional
        if "comment" not in meta:
            meta["comment"] = ""

        # Title of dataset is required
        if "title" not in meta:
            meta["title"] = ""
        if not meta["title"]:
            raise RuntimeError("Data title is missing!")

        # List of keywords is optional
        if "keywords" not in meta:
            meta["keywords"] = []

        # Description of dataset is optional
        if "description" not in meta:
            meta["description"] = ""

        # Data creation time is optional
        if "timestamp" not in meta:
            meta["timestamp"] = ""

        # Data DOI is optional
        if "doi" not in meta:
            meta["doi"] = ""

        # Data license name is optional
        if "license" not in meta:
            meta["license"] = ""

        # Store the item "meta.json"
        self["meta.json"] = meta

    def __delitem__(self, path):
        """Delete the given item."""
        # Immutable container must not be modified
        if not self.mutable:
            raise RuntimeError("Immutable container!")

        # Delete item
        if path in self:
            del self._items[path]

    def keys(self) -> typing.List[str]:
        """Return a sorted list of the full paths of all items.

        Returns:
            typing.List[str]: List of paths of Container items.
        """
        return sorted(self._items.keys())

    def values(self) -> typing.List:
        """Return a list of all item objects.

        Returns:
            typing.List: List of item objects of the Container.
        """
        return [self[k] for k in self.keys()]

    def items(self):
        """Return this container as a dictionary of item objects (key, value)
        tuples.
        """
        return {k: self[k] for k in self.keys()}

    def _hash(
        self,
        hash_object,
        item_name: str,
        content: bytes | typing.IO[bytes] | None = None,
        legacy: bool = False,
    ):
        if not legacy:
            hash_object.update(item_name.encode("utf8"))

        if isinstance(content, bytes):
            hash_object.update(content)
            return
        if content is not None and hasattr(content, "read"):
            while chunk := content.read(65536):
                hash_object.update(chunk)
            return
        if item_name in self._items and isinstance(self._items[item_name], _OnDiskFile):
            if legacy:
                raise RuntimeError(
                    "File like objects are only supported from modelVersion '1.0.1' on."
                )
            with open(self._items[item_name].path, "rb") as fp:
                while chunk := fp.read(65536):
                    hash_object.update(chunk)
            return
        if legacy:
            hash_object.update(self._items[item_name].hash().encode("ascii"))
        else:
            hash_object.update(self._items[item_name].encode())

    def hash(self):
        """Calculate and save the hash value of this container."""
        # Some attributes of content.json are excluded from the hash
        # calculation
        save = ("uuid", "created", "storageTime")
        save = {k: self["content.json"][k] for k in save}
        for key in save:
            self["content.json"][key] = None
        self["content.json"]["hash"] = None

        legacy = bool(self["content.json"]["modelVersion"] < "1.0.1")

        h = hashlib.sha256()
        for i, p in enumerate(sorted(self.items())):
            if legacy and i != 0:
                h.update(b" ")
            self._hash(h, p, legacy=legacy)
        self["content.json"]["hash"] = h.hexdigest()

        # Restore excluded attributes
        for key, value in save.items():
            self["content.json"][key] = value

        # Make container immutable
        self.mutable = False
        self["content.json"]["storageTime"] = timestamp()

    def freeze(self):
        """Calculate the hash value of this container and make it
        static. The container cannot be modified any more when this
        method was called once."""
        self["content.json"]["static"] = True
        self["content.json"]["complete"] = True
        self.hash()

    def release(self):
        """Make this container mutable. If it was immutable, this method
        will create a new UUID and initialize the attributes created,
        storageTime and modelVersion in the item "content.json".
        It will also delete an existing hash and make it a new
        container."""
        # Do nothing if the container is already mutable
        if self.mutable:
            return

        if len(self.ignore_items) > 0:
            raise RuntimeError(
                "Modifying a file that was only partially read is not supported!"
                + " Make sure 'ignore_items' is empty during initialisation."
            )
        self.mutable = True

        # Remove and initialize certain container attributes
        content = self["content.json"]
        content["static"] = False
        content["complete"] = True
        for key in ("uuid", "created", "storageTime", "hash"):
            content.pop(key, None)
        self.validate_content()

    def encode(self) -> Iterator[bytes]:
        """Encode container as ZIP package.

        Yields:
            bytes: next chunk of the generated zip file."""

        # Check/format of author ORCID
        self._norm_orcid()

        in_memory_items = {
            p: self._items[p]
            for p in self.items()
            if not isinstance(self[p], _OnDiskFile)
        }

        in_filesystem_items = {
            p: self._items[p] for p in self.items() if isinstance(self[p], _OnDiskFile)
        }

        # create FiFo queue with limited size
        queue = _StreamingQueue()

        # function to generate zip content in a seperate thread
        def _zip_generator():
            try:
                with ZipFile(
                    queue,
                    mode="w",
                    compression=self.compression,
                    compresslevel=self.compresslevel,
                ) as zfp:
                    for path in sorted(in_memory_items.keys()):
                        zfp.writestr(path, in_memory_items[path].encode())

                    for path in sorted(in_filesystem_items.keys()):
                        zfp.write(
                            in_filesystem_items[path].path,
                            arcname=path,
                            compress_type=in_filesystem_items[path].compression,
                            compresslevel=in_filesystem_items[path].compression_level,
                        )
            except Exception as e:
                queue.error = e
            finally:
                queue.close()

        # start thread to fill the queue
        thread = Thread(target=_zip_generator, daemon=True)
        thread.start()

        # yield next chunk if available
        while True:
            chunk = queue.get()
            if chunk is None:
                if queue.error is not None:
                    raise queue.error
                break

            yield chunk

    def decode(
        self,
        fp: io.RawIOBase | io.BufferedIOBase,
        ignore_items: list[str] = [],
        validate: bool = True,
        strict: bool = True,
    ):
        """Take ZIP package as file object. Read items from the
        package and store them in this object.

        Args:
            fp: File object to read from.
            ignore_items: List of file paths that are not read into memory.
            validate: If true, validate the content.
            strict: If true, validate the hash, too.
        """
        with ZipFile(fp, "r") as zfp:
            items = {p: zfp.read(p) for p in zfp.namelist() if p not in ignore_items}

            # trigger decoding content.json
            self["content.json"] = items["content.json"]
            hash = self["content.json"]["hash"]
            modelVersion = self["content.json"]["modelVersion"]

            # validate hash value, if the legacy hash function is not required
            if strict and hash and modelVersion >= "1.0.1":
                for key in ("uuid", "created", "storageTime", "hash"):
                    self["content.json"][key] = None

                h = hashlib.sha256()
                for p in sorted(zfp.namelist()):
                    if p == "content.json":
                        self._hash(h, p)
                    else:
                        with zfp.open(p) as f:
                            self._hash(h, p, content=f)

                # Check validity of hash
                if hash != h.hexdigest():
                    raise RuntimeError("Wrong hash!")
        self._store(items, validate, strict)

    def write(self, fn: str, data: bytes | None = None):
        """Write the container to a ZIP package file.

        If data is passed to the function, data will be written to the file.
        Otherwise the byte representation of the class instance will be written
        to the file, which is what you typically want.

        Args:
            fn: Filename of export file.
            data: If given, data to write to the file.
        """
        if self.mutable:
            self["content.json"]["storageTime"] = timestamp()

        if data is None:
            stream = self.encode()
        with open(fn, "wb") as fp:
            if data is not None:
                fp.write(data)
            else:
                for chunk in stream:
                    fp.write(chunk)
        self.mutable = not (
            self["content.json"]["static"] or self["content.json"]["complete"]
        )

    def _read(self, fn: str, ignore_items: list[str] = [], strict: bool = True):
        """Read a ZIP package file and store it as container in this
        object.
        """
        with open(fn, "rb") as fp:
            self.decode(fp, ignore_items, False, strict)

        self.mutable = not (
            self["content.json"]["static"] or self["content.json"]["complete"]
        )

    def upload(
        self,
        data: bytes | None = None,
        server: str | None = None,
        key: str | None = None,
    ):
        """Create a ZIP archive of the DataContainer and upload it to a server.

        If data is passed to the function, data will be written to the file.
        Otherwise the byte representation of the class instance will be written
        to the file, which is what you typically want.

        Args:
            data: If given, data to write to the file.
            server: URL of the server.
            key: API Key from the server to identify yourself.
        """
        # Server name is required and must be provided either via config
        # file, environment variable or method parameter
        if server is None:
            server = self._config["server"]
        if not server:
            raise RuntimeError("Server URL is missing!")

        # API key is required and must be provided either via config
        # file, environment variable or method parameter
        if key is None:
            key = self._config["key"]
        if not key:
            raise RuntimeError("Server API key is missing!")

        # Upload container as byte string
        if self.mutable:
            self["content.json"]["storageTime"] = timestamp()

        try:
            streamer = _MultipartRequestStreamer(data or self.encode())
            response = requests.post(
                server + "/api/datasets/",
                headers={
                    "Authorization": "Token " + key,
                    "Content-Type": f"multipart/form-data; boundary={streamer.boundary}",
                },
                data=streamer,
            )
        except Exception:
            response = None
        if response is None:
            raise ConnectionError("Connection to server %s failed!" % server)

        # HTTP status code 409 is returned when a dataset is already
        # available on the server. Static datasets require a special
        # treatment: The current dataset is replaced by the one from the
        # server.
        if response.status_code == 400:
            if self["content.json"]["static"]:
                data = json.loads(response.content.decode("UTF-8"))
                if isinstance(data, dict) and data["static"]:
                    self._download(uuid=data["id"], server=server, key=key)
                    return
            raise requests.HTTPError("400 Bad Request: Invalid container content")

        # Unauthorized access
        elif response.status_code == 403:
            raise requests.HTTPError("403 Forbidden: Unauthorized access")

        # Duplicate UUID
        elif response.status_code == 409:
            raise requests.HTTPError("409 Conflict: UUID is already existing")

        # Invalid container format
        elif response.status_code == 415:
            raise requests.HTTPError("415 Unsupported: Invalid container format")

        # Standard exception handler for other HTTP status codes
        else:
            response.raise_for_status()

        # Make container immutable
        self.mutable = not (
            self["content.json"]["static"] or self["content.json"]["complete"]
        )

    def _download(
        self, uuid, strict=True, server=None, key=None, ignore_items: list[str] = []
    ):
        # Server name is required and must be provided either via config
        # file, environment variable or method parameter
        if server is None:
            server = self._config["server"]
        if not server:
            raise RuntimeError("Server URL is missing!")

        # API key is required and must be provided either via config
        # file, environment variable or method parameter
        if key is None:
            key = self._config["key"]
        if not key:
            raise RuntimeError("Server API key is missing!")

        # Download container as byte stream from the server
        try:
            url = server + "/api/datasets/" + uuid + "/download/"
            response = requests.get(url, headers={"Authorization": "Token " + key})
        except Exception:
            response = None
        if response is None:
            raise ConnectionError("Connection to server %s failed!" % server)
        data = response.content
        fp = io.BytesIO(data)

        # Valid dataset: Store in this container
        if response.status_code == 200:
            self.decode(fp, ignore_items, False, strict)

        # Deleted dataset: Raise exception
        elif response.status_code == 204:
            raise requests.HTTPError("204 No Content: Deleted dataset")

        # Replaced dataset: Store in this container
        elif response.status_code == 301:
            self.decode(fp, ignore_items, strict)

        # Unauthorized access
        elif response.status_code == 403:
            raise requests.HTTPError("403 Forbidden: Unauthorized access")

        # Unknown dataset: Raise exception
        elif response.status_code == 404:
            raise requests.HTTPError("404 Not Found: Unknown dataset")

        # Standard exception handler for other HTTP status codes
        else:
            response.raise_for_status()

        # Make container immutable
        self.mutable = not (
            self["content.json"]["static"] or self["content.json"]["complete"]
        )

    def __str__(self):
        content = self["content.json"]
        meta = self["meta.json"]

        if content["static"]:
            ctype = "Static Container"
        elif content["complete"]:
            ctype = "Complete Container"
        else:
            ctype = "Incomplete Container"

        s = [ctype]
        ptype = content["containerType"]
        name = ptype["name"]
        if "id" in ptype:
            name = "%s %s (%s)" % (name, ptype["version"], ptype["id"])
        s.append("  type:        " + name)
        s.append("  uuid:        " + content["uuid"])
        if content["replaces"]:
            s.append("  replaces:    " + content["replaces"])
        if content["hash"]:
            s.append("  hash:        " + content["hash"])
        s.append("  created:     " + content["created"])
        s.append("  storageTime: " + content["storageTime"])
        s.append("  author:      " + meta["author"])

        return "\n".join(s)

    def _norm_orcid(self, orcid: typing.Optional[str] = None) -> None:
        """Set ORCID to normalized string if the given string is a valid ORCiD
        or to an empty string otherwise."""

        if orcid is None:
            orcid = self["meta.json"]["orcid"]

        assert isinstance(orcid, str)

        # Pick all digits and normalize
        orcid = orcid.replace("-", "").replace(" ", "").upper()
        if len(orcid) != 16:
            self["meta.json"]["orcid"] = ""
            return

        # Check if ORCID is in the correct number space.
        try:
            orcid_number = int(orcid[:-1])
        except ValueError:
            self["meta.json"]["orcid"] = ""
            return

        if not (
            (15000000 <= int(orcid_number) <= 35000000)
            or (900000000000 <= int(orcid_number) <= 900100000000)
        ):
            self["meta.json"]["orcid"] = ""
            return

        # Calculate checksum product
        r = 2
        m = 11
        try:
            product = 0
            for digit in orcid[:-1]:
                product = ((int(digit) + product) * r) % m
        except ValueError:
            self["meta.json"]["orcid"] = ""
            return

        # Calculate checksum digit
        checksum = (m + 1 - product) % m
        check = "0123456789X"[checksum]
        if orcid[-1] != check:
            self["meta.json"]["orcid"] = ""
            return

        # Valid ORCID. Set orcid variable to formatted string
        self["meta.json"]["orcid"] = (
            orcid[:4] + "-" + orcid[4:8] + "-" + orcid[8:12] + "-" + orcid[12:]
        )


class _OnDiskFile:
    """Class to represent an OnDiskFile that doesn't need to be encoded."""

    def __init__(
        self,
        path: pathlib.Path,
        compression: int | None = None,
        compression_level: int | None = None,
    ) -> None:
        self.path: pathlib.Path = path
        self.compression: int | None = compression
        self.compression_level: int | None = compression_level


class _StreamingQueue:
    def __init__(self, chunk_size: int = 65536, maxsize: int = 8):
        self.queue: Queue[bytes | None] = Queue(maxsize=maxsize)
        self.buffer: bytearray = bytearray()
        self.chunk_size: int = chunk_size
        self._closed: bool = False
        self.error: Exception | None = None

    def write(self, b: bytes) -> int:
        self.buffer += b
        if len(self.buffer) >= self.chunk_size:
            self.flush()
        return len(b)

    def flush(self):
        if self.buffer:
            try:
                self.queue.put(bytes(self.buffer), block=True, timeout=5)
            except Exception as e:
                self.error = e
            self.buffer.clear()

    def seekable(self) -> bool:
        return False

    def close(self):
        if not self._closed:
            self.flush()
            try:
                self.queue.put(None, block=True, timeout=5)
            except Exception as e:
                self.error = e
            self._closed = True

    def get(self) -> bytes | None:
        item = self.queue.get(block=True)
        if item is None and self.error:
            raise self.error
        return item


class _MultipartRequestStreamer:
    def __init__(self, data: io.RawIOBase | bytes, boundary: str | None = None) -> None:
        self.boundary: str = f"--{boundary or uuid.uuid4().hex}"
        self.data: io.RawIOBase | io.BytesIO
        if isinstance(data, bytes):
            self.data = io.BytesIO(data)
        else:
            self.data = data

        self._head_sent: bool = False
        self._file_sent: bool = False

    def __iter__(self):
        return self

    def _head(self) -> str:
        return f'--{self.boundary}\r\nContent-Disposition: form-data; name="uploadfile"; filename="dataset.zdc"\r\nContent-Type: application/octet-stream\r\n\r\n'

    def __next__(self) -> bytes:
        if not self._head_sent:
            self._head_sent = True
            return self._head().encode("utf-8")

        if not self._file_sent:
            try:
                chunk = next(self.data)
                return chunk
            except StopIteration:
                self._file_sent = True
                return f"\r\n--{self.boundary}--\r\n".encode("utf-8")
        raise StopIteration



from unittest import TestCase

from scidatacontainer import Container

from . import get_test_container


class FileIOObjectTest(TestCase):
    def test_access(self):
        a = get_test_container()

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        btext = text.encode("utf-8")

        a["data/test.exe"] = btext
        self.assertEqual(btext, a["data/test.exe"])

        a.write("test.zdc")

        b = Container(file="test.zdc", ignore_items=["data/test.exe"])

        with self.assertRaisesRegex(
            KeyError, r"Item 'data/test\.exe' was ignored while reading the file\."
        ):
            b["data/test.exe"]

    def test_release(self):
        a = get_test_container()

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        btext = text.encode("utf-8")

        a["data/test.exe"] = btext
        self.assertEqual(btext, a["data/test.exe"])

        a.write("test.zdc")
        b = Container(file="test.zdc", ignore_items=["data/test.exe"])
        with self.assertRaisesRegex(
            RuntimeError,
            r"Modifying a file that was only partially read is not supported! Make sure 'ignore_items' is empty during initialisation.",
        ):
            b.release()

    def test_load_incomplete(self):
        a = get_test_container()
        a["content.json"]["complete"] = False

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        btext = text.encode("utf-8")

        a["data/test.exe"] = btext
        self.assertEqual(btext, a["data/test.exe"])

        a.write("test.zdc")

        with self.assertRaisesRegex(
            RuntimeError,
            r"Partial loading of items only supported for immutable containers!",
        ):
            Container(file="test.zdc", ignore_items=["data/test.exe"])



from datetime import datetime, timezone

import scidatacontainer


def get_test_container():
    parameter = {
        "acquisition": {
            "acquisitionMode": "SingleFrame",
            "exposureAuto": "Off",
            "exposureMode": "Timed",
            "exposureTime": 19605.0,
            "gain": 0.0,
            "gainAuto": "Off",
            "gainMode": "Default",
            "gainSelector": "AnalogAll",
        },
        "device": {
            "family": "mvBlueFOX3",
            "id": 0,
            "product": "mvBlueFOX3-2032aG",
            "serial": "FF008343",
        },
        "format": {"height": 1544, "offsetX": 0, "offsetY": 0, "width": 2064},
    }
    data = [[float(i) for i in range(256)] for _ in range(128)]
    name = "John Doe"
    email = "john@example.com"
    org = "Example ORG"

    timestamp = scidatacontainer.timestamp()
    items = {
        "content.json": {
            "uuid": "00000000-0000-0000-0000-000000000000",
            "replaces": "00000000-0000-0000-0000-000000000001",
            "containerType": {"name": "TestType", "id": "TestID", "version": "0.1"},
            "usedSoftware": [
                {
                    "name": "testSoftware1",
                    "version": "0.2",
                },
                {
                    "name": "testSoftware2",
                    "version": "0.3",
                    "id": "TestID2",
                    "idType": "URL",
                },
            ],
            "complete": True,
            "static": False,
            "created": timestamp,
            "storageTime": timestamp,
        },
        "meta.json": {
            "author": name,
            "email": email,
            "organization": org,
            "comment": "Example comment",
            "title": "This is a sample image dataset",
            "keywords": ["keyword1", "keyword2", "keyword3"],
            "description": "Example description",
            "doi": "https://example.com/1234test1234",
            "license": "MIT",
            "timestamp": datetime(2025, 1, 1, tzinfo=timezone.utc).isoformat(
                timespec="seconds"
            ),
        },
        "meas/image.tsv": data,
        "data/parameter.json": parameter,
    }
    return scidatacontainer.Container(items=items)



from unittest import TestCase

import numpy as np

from scidatacontainer import Container, register
from scidatacontainer.filenumpy import NpyFile

from . import get_test_container


class FileNumpyTest(TestCase):
    def test_encode_decode(self):
        a = get_test_container()
        register("npx", NpyFile)

        data = np.random.randint(0, 256, (100, 100, 3))

        a["data/test.npx"] = data

        a.write("test.zdc")

        b = Container(file="test.zdc")

        self.assertTrue(np.allclose(a["data/test.npx"], b["data/test.npx"]))

    def test_hash(self):
        a = get_test_container()
        register("npx", NpyFile)

        data = np.reshape(np.arange(100 * 100 * 3), (100, 100, 3))

        a["data/test.npx"] = data
        a.freeze()

        a.write("test.zdc")

        b = Container(file="test.zdc")

        b.hash()

        self.assertEqual(
            a["content.json"]["hash"],
            b["content.json"]["hash"],
        )
        self.assertEqual(
            b["content.json"]["hash"],
            "04c7072b618f2704116a6fc172a1d06970b47e96b964a5998ab3e55b889eec88",
        )

    def test_legacy_hash(self):
        a = get_test_container()
        a["content.json"]["modelVersion"] = "1.0.0"

        register("npx", NpyFile)

        data = np.reshape(np.arange(100 * 100 * 3), (100, 100, 3))

        a["data/test.npx"] = data
        a.freeze()

        a.write("test.zdc")

        b = Container(file="test.zdc")

        b.hash()

        self.assertEqual(
            a["content.json"]["hash"],
            b["content.json"]["hash"],
        )
        self.assertEqual(
            b["content.json"]["hash"],
            "4e4ac07beaa10c8ab87f0da907f70a34e1c89918b1278dde5b75c2e4d4fd5fb6",
        )



from unittest import TestCase

from scidatacontainer import Container

from . import get_test_container


class FileBinaryTest(TestCase):
    def test_encode_decode(self):
        a = get_test_container()

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        btext = text.encode("utf-8")

        a["data/test.exe"] = btext
        self.assertEqual(btext, a["data/test.exe"])

        a.write("test.zdc")

        b = Container(file="test.zdc")

        self.assertEqual(a["data/test.exe"], b["data/test.exe"])

        self.assertEqual(b["data/test.exe"].decode("utf-8"), text)

    def test_hash(self):
        a = get_test_container()

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        btext = text.encode("utf-8")

        a["data/test.exe"] = btext

        a.freeze()
        a.write("test.zdc")

        b = Container(file="test.zdc")
        b.hash()
        self.assertEqual(a["content.json"]["hash"], b["content.json"]["hash"])
        self.assertEqual(
            b["content.json"]["hash"],
            "f66ca1f63274a613e8769902349cd6182403a45fb75df2db9afb2fe0426b9bec",
        )

    def test_legacy_hash(self):
        a = get_test_container()
        a["content.json"]["modelVersion"] = "1.0.0"

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        btext = text.encode("utf-8")

        a["data/test.exe"] = btext

        a.freeze()
        a.write("test.zdc")

        b = Container(file="test.zdc")
        b.hash()
        self.assertEqual(a["content.json"]["hash"], b["content.json"]["hash"])
        self.assertEqual(
            b["content.json"]["hash"],
            "dad3134b6b7dd2f50557ce5d891eb74faac9dcf28e03e51194d21ad4570be7b5",
        )



##########################################################################
# Copyright (c) 2023-2024 Reinhard Caspary                               #
# <reinhard.caspary@phoenixd.uni-hannover.de>                            #
# This program is free software under the terms of the MIT license.      #
##########################################################################
#
# This module provides the function load_config() to read the default
# configuration parameters
#
# Parameter           | key
# --------------------+--------
# author name         | author
# author email        | email
# author ORCiD        | orcid
# author organization | organization
# server URL          | server
# server key          | key
#
# The values of these parameters are taken either from environment
# variables or a config file. Both options are optional. Data from the
# config file overides the environment variables.
#
# The name of the environment variable of key foo is DC_FOO.
#
# The name of the config file is "$HOME\scidata.cfg" (Windows) or
# "~/.scidata" (other OS). It is expected to be a text file. Leading and
# trailing white space is ignored. Lines starting with "#" are ignored.
# The parameters are taken from lines in the form "<key> = <value>".
# Optional white space before and after the equal sign is ignored. The
# keywords are case-insensitive.
#
##########################################################################

import os
import platform


def load_config(config_path: str = None, **kwargs) -> dict:
    """Get author identity and server configuration.

    This function uses kwargs, the scidata config file and environmental
    variables as sources for each parameter. The former sources
    overriding the latter ones.

    Users may use the result of this function to inject the author
    identity when building a new container:

    config = load_config(author="John Doe", email="john@doe.com")
    dc = Container(config=config)

    Args:
        str: Path of the config file. If this is None, the default file will
             be used. This filename is only required for testing.

        kwargs: Parameter values as keyword arguments.

    Returns:
        dict: A dictionary containing information strings with keys "author",\
              "email", "orcid", "organization", "server", "key".
    """

    # Initialize config dictionary
    config = {
        "author": "",
        "email": "",
        "orcid": "",
        "organization": "",
        "server": "",
        "key": "",
    }

    # Get default values from environment variables
    for key in config:
        name = "DC_%s" % key.upper()
        if name in os.environ:
            config[key] = os.environ[name].strip()

    # Path to scidata config file
    if not config_path:
        if platform.system() == "Windows":
            config_path = os.path.join(os.path.expanduser("~"), "scidata.cfg")
        else:
            config_path = os.path.join(os.path.expanduser("~"), ".scidata")

    # Get default values from config file
    if os.path.exists(config_path):
        with open(config_path, "r", encoding="utf8") as fp:
            for line in fp.readlines():
                line = line.strip()
                if line[:1] == "#":
                    continue
                line = line.split("=", 1)
                if len(line) < 2:
                    continue
                key = line[0].strip().lower()
                if key in config:
                    config[key] = line[1].strip()

    # Use keyword arguments
    for key in config:
        if key in kwargs:
            config[key] = kwargs[key].strip()

    # Return config dictionary
    return config



import os
import unittest
from uuid import uuid4

from requests import HTTPError

from scidatacontainer import Container

from ._abstract_container_test import has_connection


class _TestDownload(unittest.TestCase):
    uuid = "00000000-0000-0000-0000-000000000000"
    test_server_url = os.getenv("SCIDATA_TEST_SERVER")
    api_key = os.getenv("SCIDATA_TEST_KEY")
    __test__ = False

    def run(self, result=None):
        if self.__test__:
            return super().run(result)

    def test_download(self):
        Container(
            uuid=self.uuid, config={"server": self.test_server_url, "key": self.api_key}
        )

    def test_replaced(self):
        uuid = "00000000-0000-0000-0000-000000000301"
        dc = Container(uuid=uuid, server=self.test_server_url, key=self.api_key)
        self.assertEqual(dc["content.json"]["uuid"], self.uuid)
        self.assertEqual(dc["content.json"]["replaces"], uuid)

    def test_no_url(self):
        with self.assertRaisesRegex(RuntimeError, "Server URL is missing!"):
            Container(uuid=self.uuid, config={"server": False, "key": self.api_key})

    def test_no_key(self):
        with self.assertRaisesRegex(RuntimeError, "Server API key is missing!"):
            Container(
                uuid=self.uuid, config={"server": self.test_server_url, "key": False}
            )

    def test_wrong_url(self):
        with self.assertRaisesRegex(
            ConnectionError, "Connection to server " + "http://localhost:54321 failed!"
        ):
            Container(
                uuid=self.uuid,
                config={"server": "http://localhost:54321", "key": self.api_key},
            )

    def test_deleted(self):
        with self.assertRaisesRegex(HTTPError, "204 No Content: Deleted dataset"):
            Container(
                uuid=self.uuid[:-3] + "204",
                server=self.test_server_url,
                key=self.api_key,
            )

    def test_unkown_uuid(self):
        with self.assertRaisesRegex(HTTPError, "404 Not Found: Unknown dataset"):
            Container(uuid=str(uuid4()), server=self.test_server_url, key=self.api_key)

    def test_invalid_key(self):
        with self.assertRaisesRegex(HTTPError, "403 Forbidden: Unauthorized access"):
            Container(
                uuid=self.uuid[:-3] + "403",
                server=self.test_server_url,
                key=self.api_key,
            )


@unittest.skipUnless(has_connection, "No server connection available.")
class TestDownload(_TestDownload):
    __test__ = True



##########################################################################
# Copyright (c) 2023 Reinhard Caspary                                    #
# <reinhard.caspary@phoenixd.uni-hannover.de>                            #
# This program is free software under the terms of the MIT license.      #
##########################################################################
#
# This module provides the base data conversion class AbstractFile and
# standard text conversion classes. All data conversion classes should
# inherit from AbstractFile and must provide three methods:
#
# encode(): Return data encoded as bytes string.
# decode(data): Decode and store given bytes string data.
# hash(): Return SHA256 hash from data as hex string.
#
# The hash implementation shoud make sure that semantically equivalent
# data results in the same hash.
#
##########################################################################

import hashlib
import json
import typing
from abc import ABC, abstractmethod

##########################################################################
# Data conversion classes


class AbstractFile(ABC):
    """Base class for converting datatypes to their file representation."""

    def __init__(self, data):
        """Constructor to create an instance of the converter class."""
        if isinstance(data, bytes):
            self.decode(data)
        else:
            self.data = data

    def hash(self) -> str:
        """Return hex digest of SHA256 hash.

        Returns:
            str: Hex digest of this object as string.
        """
        return hashlib.sha256(self.encode()).hexdigest()

    @abstractmethod
    def encode(self) -> bytes:
        """Encode the Container content to bytes. This is an abstract method
        and it needs to be overwritten by inheriting class.

        Returns:
            bytes: Byte string representation of the object.
        """
        pass  # pragma: no cover

    @abstractmethod
    def decode(self, data: bytes):
        """Decode the Container content from bytes. This is an abstract method
        and it neets to be overwritten by inheriting class.
        """
        pass  # pragma: no cover


class BinaryFile(AbstractFile):
    """Data conversion class for a binary file."""

    def encode(self) -> bytes:
        """Return byte string stored in this class.

        Returns:
            bytes: Byte string representation of the object.
        """
        return self.data

    def decode(self, data: bytes):
        """Store bytes in this class."""
        self.data = data


class TextFile(AbstractFile):
    """Data conversion class for a text file."""

    charset = "utf8"
    """charset (str): Character encoding used for translation from text to\
                      bytes."""

    def encode(self) -> bytes:
        """Encode text to bytes string.

        Returns:
            bytes: Byte string representation of the object.
        """

        return bytes(self.data, self.charset)

    def decode(self, data: bytes):
        """Decode text from given bytes string."""

        self.data = data.decode(self.charset)


class JsonFile(AbstractFile):
    """Data conversion class for a JSON file represented as Python
    dictionary."""

    indent = 4
    """indent (int): Indentation of exported JSON files."""
    charset = "utf8"
    """charset (str): Character encoding used for translation from text to\
                      bytes."""

    def sortit(self, data: typing.Union[dict, list, tuple]) -> str:
        """Return compact string representation with keys of all
        sub-dictionaries sorted.

        Args:
            data: Dictionary, list or tuple to convert to string"

        Returns:
            str: String representation of `data`
        """

        if isinstance(data, dict):
            keys = sorted(data.keys())
            data = [k + ": " + self.sortit(data[k]) for k in keys]
            data = ", ".join(data)
            data = "{" + data + "}"
            return data
        elif isinstance(data, (list, tuple)):
            data = [self.sortit(v) for v in data]
            data = ", ".join(data)
            data = "[" + data + "]"
            return data
        return repr(data)

    def hash(self) -> str:
        """Return hex digest of the SHA256 hash calculated from the
        sorted compact representation. This should result in the same
        hash for semantically equal data dictionaries.

        Returns:
            str: Hex digest of this object as string.
        """

        data = bytes(self.sortit(self.data), self.charset)
        return hashlib.sha256(data).hexdigest()

    def encode(self) -> bytes:
        """Convert dictionary to pretty string representation with
        indentation and return it as bytes string.

        Returns:
            bytes: Byte string representation of the object.
        """

        data = json.dumps(
            self.data, sort_keys=True, indent=self.indent, ensure_ascii=False
        )
        return bytes(data, self.charset)

    def decode(self, data: bytes):
        """Decode dictionary from given bytes string."""

        self.data = json.loads(data.decode(self.charset))


class TabSeparatedValuesFile(AbstractFile):
    """Data conversion class for a tab-separated value file."""

    charset = "utf8"
    """charset (str): Character encoding used for translation from a list of\
                      lists to bytes."""

    def encode(self) -> bytes:
        """Encode 2D array (list of lists) to bytes string.

        Returns:
            bytes: Byte string representation of the object.
        """
        s = "\n".join(["\t".join([str(v) for v in a]) for a in self.data])

        return bytes(s, self.charset)

    def decode(self, data: bytes):
        """Decode 2D array (list of lists) from given bytes string."""

        s = data.decode(self.charset)
        self.data = [[float(x) for x in a.split("\t")] for a in s.split("\n")]


register = [
    ("bin", BinaryFile, bytes),
    ("json", JsonFile, dict),
    ("txt", TextFile, str),
    ("log", TextFile, None),
    ("pgm", "txt", None),
    ("tsv", TabSeparatedValuesFile, None),
]



import copy
import os

from scidatacontainer import Container

from ._abstract_container_test import AbstractContainerTest


class TestStaticContainer(AbstractContainerTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.items["content.json"]["static"] = True
        cls.items_minimal["content.json"]["static"] = True

    def _compare_with_items(self, dc, check_model_version=True):
        super()._compare_with_items(dc, check_model_version)
        self._check_timestamp(dc["content.json"]["created"])

        self._check_timestamp(dc["content.json"]["storageTime"])

        self.assertTrue(dc["content.json"]["complete"])

        self.assertTrue(dc["content.json"]["static"])

    def _compare_with_minimal_items(self, dc):
        super()._compare_with_minimal_items(dc)
        self._check_timestamp(dc["content.json"]["created"])

        self._check_timestamp(dc["content.json"]["storageTime"])

        self.assertTrue(dc["content.json"]["complete"])

        self.assertTrue(dc["content.json"]["static"])

    def test_container_creation(self):
        items = copy.deepcopy(self.items)
        items["content.json"]["static"] = True
        dc = Container(items=items)

        self._compare_with_items(dc)
        self.assertTrue(dc["content.json"]["static"])
        self.assertIsNotNone(dc["content.json"]["hash"])

        self.dc = dc

    def test_container_creation_minimal(self):
        items = copy.deepcopy(self.items_minimal)
        items["content.json"]["static"] = True
        dc = Container(items=items)

        self._compare_with_minimal_items(dc)
        self.assertTrue(dc["content.json"]["static"])
        self.assertIsNotNone(dc["content.json"]["hash"])

        self.dc = dc

    def test_freeze(self):
        items = copy.deepcopy(self.items)
        items["content.json"]["static"] = False
        dc = Container(items=items)

        super()._compare_with_items(dc)
        self.assertIsNone(dc["content.json"]["hash"])
        self.assertFalse(dc["content.json"]["static"])

        dc.freeze()

        self._compare_with_items(dc)
        self.assertTrue(dc["content.json"]["static"])
        self.assertIsNotNone(dc["content.json"]["hash"])
        self.assertEqual(
            dc["content.json"]["hash"],
            "e1260e851a77e4b2d56aab0c33037a98fbdb4684b94461a9ace914bd07220dc4",
        )

        self.dc = dc

    def test_legacy_freeze(self):
        items = copy.deepcopy(self.items)
        items["content.json"]["static"] = False
        dc = Container(items=items)
        dc["content.json"]["modelVersion"] = "1.0.0"

        super()._compare_with_items(dc, check_model_version=False)
        self.assertIsNone(dc["content.json"]["hash"])
        self.assertFalse(dc["content.json"]["static"])

        dc.freeze()

        self._compare_with_items(dc, check_model_version=False)
        self.assertTrue(dc["content.json"]["static"])
        self.assertIsNotNone(dc["content.json"]["hash"])
        self.assertEqual(
            dc["content.json"]["hash"],
            "7a836281f4631bc6d17f41bf3b03dd1de4fd7fbcec08ea588c19752ab3668905",
        )

        self.items["content.json"]["modelVersion"] = "1.0.1"

    def test_freeze_minimal(self):
        items = copy.deepcopy(self.items_minimal)
        items["content.json"]["static"] = False
        dc = Container(items=items)

        super()._compare_with_minimal_items(dc)

        self.assertIsNone(dc["content.json"]["hash"])
        self.assertFalse(dc["content.json"]["static"])

        dc.freeze()

        self._compare_with_minimal_items(dc)
        self.assertTrue(dc["content.json"]["static"])
        self.assertIsNotNone(dc["content.json"]["hash"])
        self.assertEqual(
            dc["content.json"]["hash"],
            "10908bd58e451ef5c8ad1e575128fb3444e81675c434ccaf85d60d2ec8202e65",
        )

        self.dc = dc

    def test_hash_collision(self):
        self.test_freeze()

        old_hash = self.dc["content.json"]["hash"]

        self.dc.release()

        del self.dc["meas/image.tsv"]
        self.dc["meas/image3.tsv"] = self.data

        self.dc.freeze()
        self.assertNotEqual(old_hash, self.dc["content.json"]["hash"])

    def test_write(self, clean=True):
        self.test_freeze()
        self.dc.write(self.export_filename)
        if clean:
            os.remove(self.export_filename)

    def test_read(self):
        self.test_write(clean=False)
        dc = Container(file=self.export_filename)

        self._compare_with_items(dc)

        self.assertTrue(dc["content.json"]["static"])

        old_hash = dc["content.json"]["hash"]
        self.assertIsNotNone(dc["content.json"]["hash"])

        dc.hash()
        self.assertIsNotNone(dc["content.json"]["hash"])
        # new hash is equal with the old one
        self.assertEqual(old_hash, dc["content.json"]["hash"])

    def test_read_wrong_hash(self, clean=True):
        self.test_freeze()
        self.dc.content["hash"] += "ab"
        self.dc.write(self.export_filename)

        with self.assertRaises(RuntimeError) as cm:
            Container(file=self.export_filename)

        self.assertEqual(cm.exception.args[0], "Wrong hash!")

        if clean:
            os.remove(self.export_filename)

    def test_read_wrong_legacy_hash(self, clean=True):
        self.test_freeze()
        self.dc.release()
        self.dc.content["modelVersion"] = "1.0.0"
        self.dc.hash()
        self.dc.freeze()
        self.dc.content["hash"] += "ab"
        self.dc.write(self.export_filename)

        with self.assertRaises(RuntimeError) as cm:
            Container(file=self.export_filename)

        self.assertEqual(cm.exception.args[0], "Wrong hash!")

        if clean:
            os.remove(self.export_filename)

    def test_print(self):
        self.test_freeze()
        s = self.dc.__str__()

        self.assertIn("Static Container", s)
        ct = self.items["content.json"]["containerType"]
        self.assertIn(
            "type:        " + ct["name"] + " " + ct["version"] + " (" + ct["id"] + ")",
            s,
        )
        self.assertIn("uuid:        " + self.dc["content.json"]["uuid"], s)
        self.assertIn("replaces:    " + self.dc["content.json"]["replaces"], s)
        self.assertIn("hash:        " + self.dc["content.json"]["hash"], s)
        self.assertIn("created:     " + self.dc["content.json"]["created"], s)
        self.assertIn("storageTime: " + self.dc["content.json"]["storageTime"], s)
        self.assertIn("author:      " + self.dc["meta.json"]["author"], s)

    def test_immutable(self):
        self.test_freeze()
        with self.assertRaisesRegex(RuntimeError, "Immutable container!"):
            self.dc["data/test.txt"] = "Test123"



from pathlib import Path
from tempfile import NamedTemporaryFile
from unittest import TestCase
from zipfile import ZIP_DEFLATED, ZIP_STORED, ZipFile

from scidatacontainer import Container

from . import get_test_container


class FilePathTest(TestCase):
    def test_encode(self):
        a = get_test_container()

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        with NamedTemporaryFile() as tfp:
            tfp.write(text.encode("utf-8"))
            tfp.seek(0)

            a["data/test.exe"] = {"path": Path(tfp.name)}

            a.write("test.zdc")

        b = Container(file="test.zdc")
        self.assertEqual(b["data/test.exe"].decode("utf-8"), text)

    def test_invalid_file_path(self):
        a = get_test_container()

        a["data/test.exe"] = {"path": Path("/this/path/doesnt/exist")}

        with self.assertRaisesRegex(
            FileNotFoundError,
            r"\[Errno 2\] No such file or directory: '/this/path/doesnt/exist'",
        ):
            a.write("test.zdc")

    def test_hash(self):
        a = get_test_container()

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        with NamedTemporaryFile() as tfp:
            tfp.write(text.encode("utf-8"))
            tfp.seek(0)

            a["data/test.exe"] = {"path": Path(tfp.name)}

            a.freeze()
            a.write("test.zdc")

        b = Container(file="test.zdc")
        b.hash()
        self.assertEqual(a["content.json"]["hash"], b["content.json"]["hash"])
        self.assertEqual(
            b["content.json"]["hash"],
            "f66ca1f63274a613e8769902349cd6182403a45fb75df2db9afb2fe0426b9bec",
        )

    def test_changing_hash(self):
        """
        This test covers the case that a decode-encode cycle would change the data.
        This might be the case for example for json files that are not correctly formatted
        or have unsorted keys.
        """
        a = get_test_container()
        text = '{"a": "123", "b": true}'
        with NamedTemporaryFile() as tfp:
            tfp.write(text.encode("utf-8"))
            tfp.seek(0)

            a["data/test.json"] = {"path": Path(tfp.name)}

            a.freeze()
            a.write("test.zdc")

        b = Container(file="test.zdc")
        b.hash()
        self.assertNotEqual(a["content.json"]["hash"], b["content.json"]["hash"])

    def test_legacy_hash(self):
        a = get_test_container()
        a["content.json"]["modelVersion"] = "1.0.0"

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        with NamedTemporaryFile() as tfp:
            tfp.write(text.encode("utf-8"))
            tfp.seek(0)

            a["data/test.exe"] = {"path": Path(tfp.name)}

            with self.assertRaisesRegex(
                RuntimeError,
                "File like objects are only supported from modelVersion '1.0.1' on.",
            ):
                a.freeze()
                a.write("test.zdc")

    def test_compression_type(self):
        a = get_test_container()

        text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed "
        text += "do eiusmod tempor incididunt ut labore et dolore magna "
        text += "aliqua. Varius duis at consectetur lorem donec massa. Amet "
        text += "consectetur adipiscing elit ut aliquam purus sit amet. Duis "
        text += "ultricies lacus sed turpis tincidunt id aliquet risus "
        text += "feugiat. Dui id ornare arcu odio ut sem nulla pharetra."

        with NamedTemporaryFile() as tfp:
            tfp.write(text.encode("utf-8"))
            tfp.seek(0)

            a["data/test.exe"] = {"path": Path(tfp.name), "compression": ZIP_STORED}
            a.write("test.zdc")

            with open("test.zdc", "rb") as fp:
                with ZipFile(fp) as zfp:
                    self.assertEqual(
                        zfp.getinfo("data/test.exe").compress_type, ZIP_STORED
                    )

            a.release()
            a["data/test.exe"] = {"path": Path(tfp.name), "compression": ZIP_DEFLATED}
            a.write("test.zdc")

            with open("test.zdc", "rb") as fp:
                with ZipFile(fp) as zfp:
                    self.assertEqual(
                        zfp.getinfo("data/test.exe").compress_type, ZIP_DEFLATED
                    )



from unittest import TestCase

from scidatacontainer import register, AbstractFile


class DummyFile(AbstractFile):
    pass


class EmptyFileClass:
    pass


class RegisterTest(TestCase):
    def test_registration(self):
        register(".test", DummyFile, str)
        with self.assertRaisesRegex(RuntimeError, "Alias .test123:.test with" +
                                                  " default class!"):
            register(".test123", ".test", str)

        with self.assertRaises(RuntimeError) as cm:
            register(".test", EmptyFileClass, str)

        self.assertEqual(cm.exception.args[0],
                         "No method encode() in class for suffix '.test'!")



from io import BytesIO
from unittest import TestCase
from zipfile import ZipFile

import h5py
import numpy as np

from scidatacontainer import Container

from . import get_test_container


class FileHdf5Test(TestCase):
    def test_encode_decode(self):
        a = get_test_container()

        data = np.random.randint(0, 256, (10, 10, 3))

        f = h5py.File.in_memory()
        dset = f.create_dataset("hdf5_dataset", data=data)

        moredata = {
            "data1": np.random.randint(0, 256, (10, 10, 3)),
            "data2": np.random.randint(0, 256, (10, 10, 3)),
            "data3": dset,
            "attr1": "test123",
            "attr2": 123,
            "attr3": 1.23,
        }

        a["data/test.hdf5"] = data
        a["data/test2.hdf5"] = moredata

        a.write("test.zdc")

        b = Container(file="test.zdc")
        self.assertTrue(np.allclose(a["data/test.hdf5"], b["data/test.hdf5"]))
        self.assertTrue(np.allclose(moredata["data1"], b["data/test2.hdf5"]["data1"]))
        self.assertTrue(np.allclose(moredata["data2"], b["data/test2.hdf5"]["data2"]))
        self.assertTrue(np.allclose(data, b["data/test2.hdf5"]["data3"]))
        self.assertEqual(moredata["attr1"], b["data/test2.hdf5"]["attr1"])
        self.assertEqual(moredata["attr2"], b["data/test2.hdf5"]["attr2"])
        self.assertEqual(moredata["attr3"], b["data/test2.hdf5"]["attr3"])

        a.release()
        a["data/test3.hdf5"] = [data]
        with self.assertRaisesRegex(NotImplementedError, "Unknown data type list!"):
            a.write("test.zdc")

    def test_file_content(self):
        a = get_test_container()

        data = np.random.randint(0, 256, (10, 10, 3))
        moredata = {
            "data1": np.random.randint(0, 256, (10, 10, 3)),
            "data2": np.random.randint(0, 256, (10, 10, 3)),
            "attr1": "test123",
            "attr2": 123,
            "attr3": 1.23,
        }

        a["data/test.hdf5"] = data
        a["data/test2.hdf5"] = moredata

        a.write("test.zdc")

        with ZipFile("test.zdc") as zfile:
            with BytesIO(zfile.read("data/test.hdf5")) as fp:
                with h5py.File(fp) as h5file:
                    self.assertTrue(np.allclose(a["data/test.hdf5"], h5file["dataset"]))
            with BytesIO(zfile.read("data/test2.hdf5")) as fp:
                with h5py.File(fp) as h5file:
                    self.assertTrue(np.allclose(moredata["data1"], h5file["data1"]))
                    self.assertTrue(np.allclose(moredata["data2"], h5file["data2"]))
                    self.assertEqual(moredata["attr1"], h5file.attrs["attr1"])
                    self.assertEqual(moredata["attr2"], h5file.attrs["attr2"])
                    self.assertEqual(moredata["attr3"], h5file.attrs["attr3"])

    def test_hash(self):
        a = get_test_container()

        del a["meas/image.tsv"]  # remove int array that causes float precision error
        data = np.reshape(np.arange(10 * 10 * 3), (10, 10, 3))
        moredata = {
            "data1": data + 100,
            "data2": data * 10,
            "attr1": "test123",
            "attr2": 123,
            "attr3": np.float64(1.23),
            "attr4": [123, 1.24],
            "attr5": ["123", "1.24"],
            "attr6": [[123, 12], [1234, 1234]],
        }

        a["data/test.hdf5"] = data
        a["data/test2.hdf5"] = moredata
        a.freeze()
        a.write("test.zdc")

        b = Container(file="test.zdc")
        b.hash()

        self.assertEqual(a["content.json"]["hash"], b["content.json"]["hash"])
        self.assertEqual(
            b["content.json"]["hash"],
            "88843665132e891d6a332a24397970fbfe9a2f89b509193412103aefbae8e816",
        )

    def test_legacy_hash(self):
        a = get_test_container()
        a["content.json"]["modelVersion"] = "1.0.0"

        del a["meas/image.tsv"]  # remove int array that causes float precision error
        data = np.reshape(np.arange(10 * 10 * 3), (10, 10, 3))
        moredata = {
            "data1": data + 100,
            "data2": data * 10,
            "attr1": "test123",
            "attr2": 123,
            "attr3": np.float64(1.23),
            "attr4": [123, 1.24],
            "attr5": ["123", "1.24"],
            "attr6": [[123, 12], [1234, 1234]],
        }

        a["data/test.hdf5"] = data
        a["data/test2.hdf5"] = moredata
        a.freeze()
        a.write("test.zdc")

        b = Container(file="test.zdc")
        b.hash()

        self.assertEqual(a["content.json"]["hash"], b["content.json"]["hash"])
        self.assertEqual(
            b["content.json"]["hash"],
            "357bf6e2d63861cdda9239f83b23f293b3149d4d4a5283a742965199a02a7990",
        )



from __future__ import annotations  # required for |-operator in typing

__all__ = [
           "content",
           "meta",
           "validate",
           "VERSIONS_AVAILABLE",
           ]

import json
import os
from packaging import version

import jsonschema
from jsonschema.validators import Draft202012Validator, validator_for
from jsonschema.exceptions import ValidationError

_DIR = os.path.dirname(os.path.abspath(__file__))

content = {}
"""dict: Dictionary containing all available schemas for content.json
validation.

:meta hide-value:
"""

meta = {}
"""dict: Dictionary containing all available schemas for meta.json
validation.

:meta hide-value:
"""


def _translate_validation_error(exception: ValidationError,
                                filename: str):
    """Translate a jsonschema validation exception to a user friendly error
    message.

    This simply overwrites the `message` attribute of the exception.

    Args:
        exception: The exception thrown by the jsonschema library.
        filename: Filename where the problem was found.

    Raises:
        jsonschema.exceptions.ValidationError: Modified exception.
    """
    if "errorMessage" in exception.validator_value:
        exception.message = exception.validator_value["errorMessage"]
    elif exception.validator == "required":
        exception.message = exception.message + " in " + filename + ".json."
    elif exception.validator == "format":
        exception.message = "Value of '" + exception.path[0] + "' in " +\
                            filename + ".json has the wrong format: " +\
                            exception.message + "."
    elif exception.validator == "type":
        exception.message = "Value of '" + exception.path[0] + "' in " +\
                            filename + ".json has the wrong type: " +\
                            exception.message + "."
    elif exception.validator == "pattern":
        exception.message = "Value of '" + exception.path[0] + "' in " +\
                            filename + ".json has the wrong format: "
        if "patternErrorMessage" in exception.schema:
            exception.message += exception.schema["patternErrorMessage"]
        else:
            exception.message += exception.message + "."

    raise exception


def _guess_schema(instance: dict) -> (dict, str):
    """Guess the schema type and version from an instance.

    The schema is guessed by finding the schema which throws the lowest amount
    of errors during validation.

    Args:
        instance: Instance with unknown schema.

    Returns:
        schema: The schema that fits the best
        schema_name: schema type of the best match ('meta' or 'content')
    """
    schema = None
    schema_name = None
    min_errors = 1e30

    for key, s in content.items():
        cls = validator_for(s)
        num_errors = 0
        for _ in cls(s).iter_errors(instance):
            num_errors += 1
        if num_errors < min_errors:
            min_errors = num_errors
            schema = s
            schema_name = "content"

    for key, s in meta.items():
        cls = validator_for(s)
        num_errors = 0
        for _ in cls(s).iter_errors(instance):
            num_errors += 1
        if num_errors < min_errors:
            min_errors = num_errors
            schema = s
            schema_name = "meta"

    return schema, schema_name


def validate(instance: dict,
             schema: dict | None = None,
             schema_name: str | None = None,
             translate: bool = True,
             check_format: bool = True):
    """Validate if a dictionary fulfills the requirements of a
        SciDataContainer.

        If schema or schema_name is missing the values are guessed by finding
        the best overlap with one of the schemas.

        Args:
            instance: meta data dictionary.
            schema: schema used for validation.
            schema_name: "content" or "meta" used for more verbose error
                messages.
            translate: If true, replace the error message with more user
                friendly messages.
            check_format: If true, check the format of strings if they match
                the required format (e.g. date-time, email, ...).

        Raises:
            jsonschema.exceptions.ValidationError: Exception with details about
                the problem of the validated dictionary.
    """

    if check_format:
        format_checker = Draft202012Validator.FORMAT_CHECKER
    else:
        format_checker = None

    if (not schema) or (not schema_name) or\
            (schema_name not in ["meta", "content"]):
        schema, schema_name = _guess_schema(instance)

    try:
        jsonschema.validate(schema=schema,
                            instance=instance,
                            format_checker=format_checker)
    except ValidationError as e:
        if translate:
            _translate_validation_error(e, schema_name)
        else:
            raise


VERSIONS_AVAILABLE = []

for filename in os.listdir(_DIR):
    if filename.startswith("SciDataContainer.content."):
        modelVersion = filename[len("SciDataContainer.content."):]
        modelVersion = modelVersion[:-len(".schema.json")]
        if modelVersion not in VERSIONS_AVAILABLE:
            VERSIONS_AVAILABLE.append(modelVersion)
        with open(_DIR + "/" + filename, 'r') as content_file:
            content_json = content_file.read()
            content[modelVersion] = json.loads(content_json)
            Draft202012Validator.check_schema(content[modelVersion])

    elif filename.startswith("SciDataContainer.meta."):
        modelVersion = filename[len("SciDataContainer.meta."):]
        modelVersion = modelVersion[:-len(".schema.json")]
        if modelVersion not in VERSIONS_AVAILABLE:
            VERSIONS_AVAILABLE.append(modelVersion)
        with open(_DIR + "/" + filename, 'r') as meta_file:
            meta_json = meta_file.read()
            meta[modelVersion] = json.loads(meta_json)
            Draft202012Validator.check_schema(meta[modelVersion])

VERSIONS_AVAILABLE.sort(key=version.parse)



import os
from unittest import TestCase

from scidatacontainer.config import load_config


class ConfigTest(TestCase):
    def check_cfg_empty(self):
        cfg = load_config("./thisfiledoesntexist.cfg")

        self.assertEqual(cfg["author"], "")
        self.assertEqual(cfg["email"], "")
        self.assertEqual(cfg["server"], "")
        self.assertEqual(cfg["key"], "")

    def test_envvars(self):
        self.check_cfg_empty()

        os.environ["DC_AUTHOR"] = "Maxi Mustermann"
        os.environ["DC_EMAIL"] = "maxi@example.com"
        os.environ["DC_SERVER"] = "http://example.com"
        os.environ["DC_KEY"] = "SECRETKEY"

        cfg = load_config("./thisfiledoesntexist.cfg")

        self.assertEqual(cfg["author"], "Maxi Mustermann")
        self.assertEqual(cfg["email"], "maxi@example.com")
        self.assertEqual(cfg["server"], "http://example.com")
        self.assertEqual(cfg["key"], "SECRETKEY")

    def test_cfg_file(self):
        self.check_cfg_empty()
        tmp_cfg = "./tmp_cfg"

        with open(tmp_cfg, 'w') as f:
            f.write("# Test comment = should be skipped\n")
            f.write("author=Maxi Mustermann\n")
            f.write("email=maxi@example.com\n")
            f.write("server=http://example.com\n")
            f.write("key=\n")
            f.write("key=SECRETKEY\n")
            f.write("key invalid line and should be ignored\n")

        cfg = load_config(tmp_cfg)

        self.assertEqual(cfg["author"], "Maxi Mustermann")
        self.assertEqual(cfg["email"], "maxi@example.com")
        self.assertEqual(cfg["server"], "http://example.com")
        self.assertEqual(cfg["key"], "SECRETKEY")

        os.remove(tmp_cfg)



import uuid
from abc import ABC
from datetime import datetime, timedelta, timezone
from unittest import TestCase

import requests

import scidatacontainer


class AbstractContainerTest(ABC, TestCase):
    @classmethod
    def setUpClass(cls):
        # Dummy parameters: a dict
        cls.parameter = {
            "acquisition": {
                "acquisitionMode": "SingleFrame",
                "exposureAuto": "Off",
                "exposureMode": "Timed",
                "exposureTime": 19605.0,
                "gain": 0.0,
                "gainAuto": "Off",
                "gainMode": "Default",
                "gainSelector": "AnalogAll",
            },
            "device": {
                "family": "mvBlueFOX3",
                "id": 0,
                "product": "mvBlueFOX3-2032aG",
                "serial": "FF008343",
            },
            "format": {"height": 1544, "offsetX": 0, "offsetY": 0, "width": 2064},
        }

        cls.data = [[float(i) for i in range(256)] for j in range(128)]
        cls.name = "John Doe"
        cls.email = "john@example.com"
        cls.org = "Example ORG"

        cls.timestamp = scidatacontainer.timestamp()
        cls.items = {
            "content.json": {
                "uuid": str(uuid.uuid4()),
                "replaces": "b185cc38-1c62-404b-8373-89a4605845ca",
                "containerType": {"name": "TestType", "id": "TestID", "version": "0.1"},
                "usedSoftware": [
                    {
                        "name": "testSoftware1",
                        "version": "0.2",
                    },
                    {
                        "name": "testSoftware2",
                        "version": "0.3",
                        "id": "TestID2",
                        "idType": "URL",
                    },
                ],
                "complete": True,
                "static": False,
                "created": cls.timestamp,
                "storageTime": cls.timestamp,
            },
            "meta.json": {
                "author": cls.name,
                "email": cls.email,
                "organization": cls.org,
                "comment": "Example comment",
                "title": "This is a sample image dataset",
                "keywords": ["keyword1", "keyword2", "keyword3"],
                "description": "Example description",
                "doi": "https://example.com/1234abc1234",
                "license": "MIT",
                "timestamp": datetime(2025, 1, 1, tzinfo=timezone.utc).isoformat(
                    timespec="seconds"
                ),
            },
            "meas/image.tsv": cls.data,
            "data/parameter.json": cls.parameter,
        }

        cls.items_minimal = {
            "content.json": {
                "containerType": {"name": "TestType"},
                "usedSoftware": [],
            },
            "meta.json": {
                "author": cls.name,
                "email": cls.email,
                "title": "This is a sample image dataset",
            },
            "meas/image.tsv": cls.data,
            "data/parameter.json": cls.parameter,
        }
        cls.export_filename = "test.zdc"

    def _compare_with_items(self, dc, check_model_version=True):
        items = self.items
        self.assertEqual(dc["content.json"]["uuid"], items["content.json"]["uuid"])

        self.assertEqual(dc["content.json"]["uuid"], items["content.json"]["uuid"])

        self.assertEqual(
            items["content.json"]["containerType"], dc["content.json"]["containerType"]
        )

        self.assertEqual(
            items["content.json"]["usedSoftware"], dc["content.json"]["usedSoftware"]
        )

        if check_model_version:
            self.assertEqual(
                dc["content.json"]["modelVersion"], scidatacontainer.modelVersion
            )

        self.assertEqual(items["meta.json"]["author"], dc["meta.json"]["author"])

        self.assertEqual(items["meta.json"]["email"], dc["meta.json"]["email"])

        self.assertEqual(
            items["meta.json"]["organization"], dc["meta.json"]["organization"]
        )

        self.assertEqual(items["meta.json"]["comment"], dc["meta.json"]["comment"])

        self.assertEqual(items["meta.json"]["title"], dc["meta.json"]["title"])

        self.assertEqual(items["meta.json"]["keywords"], dc["meta.json"]["keywords"])

        self.assertEqual(
            items["meta.json"]["description"], dc["meta.json"]["description"]
        )

        self.assertEqual(
            datetime(2025, 1, 1, tzinfo=timezone.utc).isoformat(timespec="seconds"),
            dc["meta.json"]["timestamp"],
        )

        self.assertEqual(items["meta.json"]["doi"], dc["meta.json"]["doi"])

        self.assertEqual(items["meta.json"]["license"], dc["meta.json"]["license"])

        self.assertEqual(dc["meas/image.tsv"], self.data)

        self.assertEqual(dc["data/parameter.json"], self.parameter)

    def _compare_with_minimal_items(self, dc):
        items = self.items_minimal
        # check if uuid is a valid uuid
        self.assertTrue(isinstance(dc["content.json"]["uuid"], str))
        self.assertEqual(
            dc["content.json"]["uuid"], str(uuid.UUID(dc["content.json"]["uuid"]))
        )

        self.assertIsNone(dc["content.json"]["replaces"])

        self.assertEqual(
            items["content.json"]["containerType"], dc["content.json"]["containerType"]
        )

        self.assertEqual(
            items["content.json"]["usedSoftware"], dc["content.json"]["usedSoftware"]
        )

        self.assertEqual(
            dc["content.json"]["modelVersion"], scidatacontainer.modelVersion
        )

        self.assertEqual(items["meta.json"]["author"], dc["meta.json"]["author"])

        self.assertEqual(items["meta.json"]["email"], dc["meta.json"]["email"])

        self.assertEqual("", dc["meta.json"]["organization"])

        self.assertEqual("", dc["meta.json"]["comment"])

        self.assertEqual(items["meta.json"]["title"], dc["meta.json"]["title"])

        self.assertEqual([], dc["meta.json"]["keywords"])

        self.assertEqual("", dc["meta.json"]["description"])

        self.assertEqual("", dc["meta.json"]["timestamp"])

        self.assertEqual("", dc["meta.json"]["doi"])

        self.assertEqual("", dc["meta.json"]["license"])

        self.assertEqual(dc["meas/image.tsv"], self.data)

        self.assertEqual(dc["data/parameter.json"], self.parameter)

    def _check_timestamp(self, timestamp):
        t1 = datetime.fromisoformat(timestamp)
        t2 = datetime.fromisoformat(self.timestamp)
        self.assertTrue(abs(t1 - t2) < timedelta(seconds=10))


def _check_server_connection():
    cfg = scidatacontainer.config.load_config()
    if (cfg["server"] == "") or (cfg["key"] == ""):
        return False

    try:
        response = requests.get(
            cfg["server"] + "/api/datasets/",
            headers={"Authorization": "Token " + cfg["key"]},
        )
    except requests.exceptions.ConnectionError:
        return False
    return response.status_code == 200


has_connection = _check_server_connection()
